from __future__ import annotations

from typing import Any, Callable

from Live import *
from Live.Application import *
from Live.Base import *
from Live.Browser import *
from Live.CcControlDevice import *
from Live.ChainMixerDevice import *
from Live.Clip import *
from Live.ClipSlot import *
from Live.CompressorDevice import *
from Live.Conversions import *
from Live.Device import *
from Live.DeviceIO import *
from Live.DeviceParameter import *
from Live.DriftDevice import *
from Live.DrumChain import *
from Live.DrumPad import *
from Live.Eq8Device import *
from Live.Groove import *
from Live.GroovePool import *
from Live.HybridReverbDevice import *
from Live.Listener import *
from Live.LomObject import *
from Live.MaxDevice import *
from Live.MeldDevice import *
from Live.MidiMap import *
from Live.MixerDevice import *
from Live.PluginDevice import *
from Live.RackDevice import *
from Live.RoarDevice import *
from Live.Sample import *
from Live.Scene import *
from Live.ShifterDevice import *
from Live.SimplerDevice import *
from Live.Song import *
from Live.SpectralResonatorDevice import *
from Live.Track import *
from Live.WavetableDevice import *


class Chain:
    """
    This class represents a group device chain in Live.
    """

    def add_color_index_listener(self, listener: Callable) -> None:
        """
        add_color_index_listener( (Chain)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "color_index" has changed.

            C++ signature :
                void add_color_index_listener(TChainPyHandle,boost::python::api::object)
        """

    def add_color_listener(self, listener: Callable) -> None:
        """
        add_color_listener( (Chain)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "color" has changed.

            C++ signature :
                void add_color_listener(TChainPyHandle,boost::python::api::object)
        """

    def add_devices_listener(self, listener: Callable) -> None:
        """
        add_devices_listener( (Chain)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "devices" has changed.

            C++ signature :
                void add_devices_listener(TChainPyHandle,boost::python::api::object)
        """

    def add_is_auto_colored_listener(self, listener: Callable) -> None:
        """
        add_is_auto_colored_listener( (Chain)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "is_auto_colored" has changed.

            C++ signature :
                void add_is_auto_colored_listener(TChainPyHandle,boost::python::api::object)
        """

    def add_mute_listener(self, listener: Callable) -> None:
        """
        add_mute_listener( (Chain)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "mute" has changed.

            C++ signature :
                void add_mute_listener(TChainPyHandle,boost::python::api::object)
        """

    def add_muted_via_solo_listener(self, listener: Callable) -> None:
        """
        add_muted_via_solo_listener( (Chain)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "muted_via_solo" has changed.

            C++ signature :
                void add_muted_via_solo_listener(TChainPyHandle,boost::python::api::object)
        """

    def add_name_listener(self, listener: Callable) -> None:
        """
        add_name_listener( (Chain)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "name" has changed.

            C++ signature :
                void add_name_listener(TChainPyHandle,boost::python::api::object)
        """

    def add_solo_listener(self, listener: Callable) -> None:
        """
        add_solo_listener( (Chain)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "solo" has changed.

            C++ signature :
                void add_solo_listener(TChainPyHandle,boost::python::api::object)
        """

    def color_has_listener(self, listener: Callable) -> bool:
        """
        color_has_listener( (Chain)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "color".

            C++ signature :
                bool color_has_listener(TChainPyHandle,boost::python::api::object)
        """

    def color_index_has_listener(self, listener: Callable) -> bool:
        """
        color_index_has_listener( (Chain)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "color_index".

            C++ signature :
                bool color_index_has_listener(TChainPyHandle,boost::python::api::object)
        """

    def delete_device(self, arg2: int) -> None:
        """
        delete_device( (Chain)arg1, (int)arg2) -> None :
            Remove a device identified by its index from the chain. Throws runtime error if bad index.


            C++ signature :
                void delete_device(TChainPyHandle,int)
        """

    def devices_has_listener(self, listener: Callable) -> bool:
        """
        devices_has_listener( (Chain)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "devices".

            C++ signature :
                bool devices_has_listener(TChainPyHandle,boost::python::api::object)
        """

    def is_auto_colored_has_listener(self, listener: Callable) -> bool:
        """
        is_auto_colored_has_listener( (Chain)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "is_auto_colored".

            C++ signature :
                bool is_auto_colored_has_listener(TChainPyHandle,boost::python::api::object)
        """

    def mute_has_listener(self, listener: Callable) -> bool:
        """
        mute_has_listener( (Chain)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "mute".

            C++ signature :
                bool mute_has_listener(TChainPyHandle,boost::python::api::object)
        """

    def muted_via_solo_has_listener(self, listener: Callable) -> bool:
        """
        muted_via_solo_has_listener( (Chain)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "muted_via_solo".

            C++ signature :
                bool muted_via_solo_has_listener(TChainPyHandle,boost::python::api::object)
        """

    def name_has_listener(self, listener: Callable) -> bool:
        """
        name_has_listener( (Chain)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "name".

            C++ signature :
                bool name_has_listener(TChainPyHandle,boost::python::api::object)
        """

    def remove_color_index_listener(self, listener: Callable) -> None:
        """
        remove_color_index_listener( (Chain)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "color_index".

            C++ signature :
                void remove_color_index_listener(TChainPyHandle,boost::python::api::object)
        """

    def remove_color_listener(self, listener: Callable) -> None:
        """
        remove_color_listener( (Chain)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "color".

            C++ signature :
                void remove_color_listener(TChainPyHandle,boost::python::api::object)
        """

    def remove_devices_listener(self, listener: Callable) -> None:
        """
        remove_devices_listener( (Chain)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "devices".

            C++ signature :
                void remove_devices_listener(TChainPyHandle,boost::python::api::object)
        """

    def remove_is_auto_colored_listener(self, listener: Callable) -> None:
        """
        remove_is_auto_colored_listener( (Chain)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "is_auto_colored".

            C++ signature :
                void remove_is_auto_colored_listener(TChainPyHandle,boost::python::api::object)
        """

    def remove_mute_listener(self, listener: Callable) -> None:
        """
        remove_mute_listener( (Chain)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "mute".

            C++ signature :
                void remove_mute_listener(TChainPyHandle,boost::python::api::object)
        """

    def remove_muted_via_solo_listener(self, listener: Callable) -> None:
        """
        remove_muted_via_solo_listener( (Chain)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "muted_via_solo".

            C++ signature :
                void remove_muted_via_solo_listener(TChainPyHandle,boost::python::api::object)
        """

    def remove_name_listener(self, listener: Callable) -> None:
        """
        remove_name_listener( (Chain)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "name".

            C++ signature :
                void remove_name_listener(TChainPyHandle,boost::python::api::object)
        """

    def remove_solo_listener(self, listener: Callable) -> None:
        """
        remove_solo_listener( (Chain)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "solo".

            C++ signature :
                void remove_solo_listener(TChainPyHandle,boost::python::api::object)
        """

    def solo_has_listener(self, listener: Callable) -> bool:
        """
        solo_has_listener( (Chain)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "solo".

            C++ signature :
                bool solo_has_listener(TChainPyHandle,boost::python::api::object)
        """

    @property
    def canonical_parent(self) -> Any:
        """
        Get the canonical parent of the chain.
        """

    @property
    def color(self) -> Any:
        """
        Access the color index of the Chain.
        """

    @property
    def color_index(self) -> Any:
        """
        Access the color index of the Chain.
        """

    @property
    def devices(self) -> list[Device]:
        """
        Return const access to all available Devices that are present in the chains
        """

    @property
    def has_audio_input(self) -> bool:
        """
        return True, if this Chain can be feed with an Audio signal. This is
        true for all Audio Chains.
        """

    @property
    def has_audio_output(self) -> bool:
        """
        return True, if this Chain sends out an Audio signal. This is
        true for all Audio Chains, and MIDI chains with an Instrument.
        """

    @property
    def has_midi_input(self) -> bool:
        """
        return True, if this Chain can be feed with an Audio signal. This is
        true for all MIDI Chains.
        """

    @property
    def has_midi_output(self) -> bool:
        """
        return True, if this Chain sends out MIDI events. This is
        true for all MIDI Chains with no Instruments.
        """

    @property
    def is_auto_colored(self) -> bool:
        """
        Get/set access to the auto color flag of the Chain.
        If True, the Chain will always have the same color as the containing
        Track or Chain.
        """

    @is_auto_colored.setter
    def is_auto_colored(self, value: bool) -> None:
        pass

    @property
    def mixer_device(self) -> Device:
        """
        Return access to the mixer device that holds the chain's mixer parameters:
        the Volume, Pan, and Sendamounts.
        """

    @property
    def mute(self) -> Any:
        """
        Mute/unmute the chain.
        """

    @property
    def muted_via_solo(self) -> Any:
        """
        Return const access to whether this chain is muted due to some other chain
        being soloed.
        """

    @property
    def name(self) -> str:
        """
        Read/write access to the name of the Chain, as visible in the track header.
        """

    @name.setter
    def name(self, value: str) -> None:
        pass

    @property
    def solo(self) -> Any:
        """
        Get/Set the solo status of the chain. Note that this will not disable the
        solo state of any other Chain in the same rack. If you want exclusive solo,
        you have to disable the solo state of the other Chains manually.
        """

    @solo.setter
    def solo(self, value: Any) -> None:
        pass



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/components/view_control.py
# Compiled at: 2023-12-21 15:35:35
# Size of source mod 2**32: 7308 bytes
from functools import partial
from typing import cast

from ...base import EventObject, ObservablePropertyAlias, clamp, depends
from ...live import all_visible_tracks, scene_index, track_index
from .. import Component
from ..display import Renderable
from . import Scrollable, ScrollComponent


class NotifyingViewScroller(Scrollable, EventObject):
    __events__ = ("scrolled",)

    def __init__(self, song, scroll_scenes=False, page_size=1, *a, **k):
        (super().__init__)(*a, **k)
        self.song = song
        self.page_size = page_size
        can_scroll = self._can_scroll_scenes if scroll_scenes else self._can_scroll_tracks
        self.can_scroll_up = partial(can_scroll, -1)
        self.can_scroll_down = partial(can_scroll, 1)
        self._do_scroll = self._scroll_scenes if scroll_scenes else self._scroll_tracks

    def scroll_up(self):
        if self.can_scroll_up():
            self._do_scroll(-self.page_size)

    def scroll_down(self):
        if self.can_scroll_down():
            self._do_scroll(self.page_size)

    @staticmethod
    def _can_scroll_tracks(delta):
        tracks = all_visible_tracks()
        return track_index(track_list=tracks) + delta in range(len(tracks))

    def _can_scroll_scenes(self, delta):
        return scene_index() + delta in range(len(self.song.scenes))

    def _scroll_tracks(self, delta):
        tracks = all_visible_tracks()
        new_index = track_index(track_list=tracks) + delta
        self.song.view.selected_track = tracks[clamp(new_index, 0, len(tracks) - 1)]
        self.notify_scrolled()

    def _scroll_scenes(self, delta):
        scenes = list(self.song.scenes)
        new_index = scene_index() + delta
        self.song.view.selected_scene = scenes[clamp(new_index, 0, len(scenes) - 1)]
        self.notify_scrolled()


class ViewControlComponent(Component, Renderable):
    __events__ = ("track_selection_scrolled", "scene_selection_scrolled")

    @depends(session_ring=None)
    def __init__(self, name="View_Control", session_ring=None, *a, **k):
        (super().__init__)(a, name=name, **k)
        self._session_ring = session_ring

        def scroller_factory(scroll_scenes=False, **k):
            scroller_type = "Scene" if scroll_scenes else "Track"
            scroller = NotifyingViewScroller(self.song, scroll_scenes=scroll_scenes, **k)
            self.register_disconnectable(
                ObservablePropertyAlias(
                    self,
                    property_host=scroller,
                    property_name="scrolled",
                    alias_name=(f"{scroller_type.lower()}_selection_scrolled"),
                ),
            )
            return ScrollComponent(scroller, parent=self, scroll_skin_name=(f"ViewControl.{scroller_type}"))

        self._scroll_tracks = scroller_factory()
        self._page_tracks = scroller_factory(page_size=(self._session_ring.num_tracks))
        self._scroll_scenes = scroller_factory(scroll_scenes=True)
        self._page_scenes = scroller_factory(scroll_scenes=True, page_size=(self._session_ring.num_scenes))
        song = self.song
        view = song.view
        self.register_slot(self._session_ring, self._update_track_scrollers, "tracks")
        self.register_slot(song, self._update_track_scrollers, "visible_tracks")
        self.register_slot(song, self._update_track_scrollers, "return_tracks")
        self.register_slot(view, self._update_track_scrollers, "selected_track")
        self.register_slot(song, self._update_scene_scrollers, "scenes")
        self.register_slot(view, self._update_scene_scrollers, "selected_scene")

        def notify_tracks_scrolled():
            self.notify(self.notifications.Track.select, cast(str, self.song.view.selected_track.name))

        self.register_slot(self, notify_tracks_scrolled, "track_selection_scrolled")

    def set_next_track_button(self, button):
        self._scroll_tracks.set_scroll_down_button(button)

    def set_prev_track_button(self, button):
        self._scroll_tracks.set_scroll_up_button(button)

    def set_next_track_page_button(self, button):
        self._page_tracks.set_scroll_down_button(button)

    def set_prev_track_page_button(self, button):
        self._page_tracks.set_scroll_up_button(button)

    def set_track_encoder(self, control):
        self._scroll_tracks.set_scroll_encoder(control)

    def set_track_page_encoder(self, control):
        self._page_tracks.set_scroll_encoder(control)

    def set_next_scene_button(self, button):
        self._scroll_scenes.set_scroll_down_button(button)

    def set_prev_scene_button(self, button):
        self._scroll_scenes.set_scroll_up_button(button)

    def set_next_scene_page_button(self, button):
        self._page_scenes.set_scroll_down_button(button)

    def set_prev_scene_page_button(self, button):
        self._page_scenes.set_scroll_up_button(button)

    def set_scene_encoder(self, control):
        self._scroll_scenes.set_scroll_encoder(control)

    def set_scene_page_encoder(self, control):
        self._page_scenes.set_scroll_encoder(control)

    def _update_track_scrollers(self):
        self._scroll_tracks.update()
        self._page_tracks.update()

    def _update_scene_scrollers(self):
        self._scroll_scenes.update()
        self._page_scenes.update()



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/elements/__init__.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 2682 bytes
from .button import ButtonElement, ButtonElementMixin, DummyUndoStepHandler
from .button_matrix import ButtonMatrixElement
from .button_slider import ButtonSliderElement
from .color import (
    AnimatedColor,
    Color,
    DynamicColorBase,
    SelectedClipColorFactory,
    SelectedTrackColor,
    SelectedTrackColorFactory,
    SysexRGBColor,
    to_midi_value,
)
from .combo import (
    ComboElement,
    DoublePressContext,
    DoublePressElement,
    EventElement,
    MultiElement,
    ToggleElement,
    WrapperElement,
)
from .display_data_source import DisplayDataSource, adjust_string, adjust_string_crop
from .encoder import EncoderElement, FineGrainWithModifierEncoderElement, TouchEncoderElement, TouchEncoderElementBase
from .full_velocity_element import FullVelocityElement, NullFullVelocity
from .logical_display_segment import LogicalDisplaySegment
from .optional import ChoosingElement, OptionalElement
from .physical_display import (
    DisplayElement,
    DisplayError,
    DisplaySegmentationError,
    PhysicalDisplayElement,
    SubDisplayElement,
)
from .playhead_element import NullPlayhead, PlayheadElement
from .proxy_element import ProxyElement
from .slider import SliderElement
from .sysex_element import ColorSysexElement, SysexElement
from .velocity_levels_element import NullVelocityLevels, VelocityLevelsElement

__all__ = (
    "AnimatedColor",
    "ButtonElement",
    "ButtonElementMixin",
    "ButtonValue",
    "Color",
    "ColorSysexElement",
    "DummyUndoStepHandler",
    "DynamicColorBase",
    "OFF_VALUE",
    "ON_VALUE",
    "ButtonMatrixElement",
    "ButtonSliderElement",
    "ComboElement",
    "DoublePressContext",
    "DoublePressElement",
    "EventElement",
    "FullVelocityElement",
    "MultiElement",
    "ToggleElement",
    "WrapperElement",
    "adjust_string",
    "adjust_string_crop",
    "DisplayDataSource",
    "EncoderElement",
    "FineGrainWithModifierEncoderElement",
    "TouchEncoderElement",
    "TouchEncoderElementBase",
    "LogicalDisplaySegment",
    "ChoosingElement",
    "OptionalElement",
    "DisplayElement",
    "DisplayError",
    "DisplaySegmentationError",
    "NullFullVelocity",
    "NullPlayhead",
    "NullVelocityLevels",
    "PhysicalDisplayElement",
    "PlayheadElement",
    "ProxyElement",
    "SubDisplayElement",
    "SelectedTrackColor",
    "SelectedTrackColorFactory",
    "SelectedClipColorFactory",
    "SliderElement",
    "SysexElement",
    "SysexRGBColor",
    "to_midi_value",
    "VelocityLevelsElement",
)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/resource.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 9672 bytes
from functools import partial, reduce

from ..base import NamedTuple, Proxy, first, index_if, nop

DEFAULT_PRIORITY = 0


class Resource:
    def grab(self, client, *a, **k):
        raise NotImplementedError

    def release(self, client):
        raise NotImplementedError

    def get_owner(self):
        raise NotImplementedError

    owner = property(
        lambda self: self.get_owner(),
    )


class CompoundResource(Resource):
    def __init__(self, first_resource=None, second_resource=None, *a, **k):
        (super().__init__)(*a, **k)
        self._first_resource = first_resource
        self._second_resource = second_resource

    def grab(self, client, *a, **k):
        if (self._first_resource.grab)(client, *a, **k):
            if (self._second_resource.grab)(client, *a, **k):
                pass
            else:
                self._first_resource.release(client)
        return self.owner == client

    def release(self, client):
        if client == self.owner:
            self._second_resource.release(client)
            self._first_resource.release(client)
            return True
        return False

    def get_owner(self):
        return self._first_resource.owner or self._second_resource.owner

    @property
    def first(self):
        return self._first_resource

    @property
    def second(self):
        return self._second_resource


def compose_resources(*resources):
    return reduce(CompoundResource, resources)


class ExclusiveResource(Resource):
    def __init__(self, on_received_callback=None, on_lost_callback=None, *a, **k):
        (super().__init__)(*a, **k)
        self._owner = None
        if on_received_callback:
            self.on_received = on_received_callback
        if on_lost_callback:
            self.on_lost = on_lost_callback

    def grab(self, client, *a, **k):
        if self._owner is None:
            (self.on_received)(client, *a, **k)
            self._owner = client
        return self._owner == client

    def release(self, client):
        if client == self._owner:
            self._owner = None
            self.on_lost(client)
            return True
        return False

    def get_owner(self):
        return self._owner

    def on_received(self, client, *a, **k):
        raise NotImplementedError("Override or pass callback")

    def on_lost(self, client):
        raise NotImplementedError("Override or pass callback")


class SharedResource(Resource):
    def __init__(self, on_received_callback=None, on_lost_callback=None, *a, **k):
        (super().__init__)(*a, **k)
        if on_received_callback:
            self.on_received = on_received_callback
        if on_lost_callback:
            self.on_lost = on_lost_callback
        self._clients = set()

    def grab(self, client, *a, **k):
        (self.on_received)(client, *a, **k)
        self._clients.add(client)
        return True

    def release(self, client):
        if client in self._clients:
            self.on_lost(client)
            self._clients.remove(client)
            for client in self._clients:
                self.on_received(client)

            return True
        return False

    def get_owner(self):
        pass

    def on_received(self, client, *a, **k):
        raise NotImplementedError("Override or pass callback")

    def on_lost(self, client):
        raise NotImplementedError("Override or pass callback")


class StackingResource(Resource):
    def __init__(self, on_received_callback=None, on_lost_callback=None, *a, **k):
        (super().__init__)(*a, **k)
        self._clients = []
        self._owners = set()
        if on_received_callback:
            self.on_received = on_received_callback
        if on_lost_callback:
            self.on_lost = on_lost_callback

    def grab(self, client, priority=None):
        if priority is None:
            priority = DEFAULT_PRIORITY
        old_owners = self._owners
        self._remove_client(client)
        self._add_client(client, priority)
        new_owners = self._actual_owners()
        if new_owners != old_owners:
            self._owners = new_owners
            self._on_lost_set(set(old_owners) - set(new_owners))
            self._on_received_set(new_owners)
        return True

    def _on_lost_set(self, clients):
        for client in clients:
            self.on_lost(client)

    def _on_received_set(self, clients):
        for client in clients:
            self.on_received(client)

    def release(self, client):
        old_owners = self._owners
        result = self._remove_client(client)
        new_owners = self._actual_owners()
        if new_owners != old_owners:
            self._owners = new_owners
            self._on_lost_set(set(old_owners) - set(new_owners))
            self._on_received_set(new_owners)
        return result

    def release_all(self):
        for client, _ in list(self._clients):
            self.release(client)

    def _add_client(self, client, priority):
        idx = index_if(lambda client_and_priority: client_and_priority[1] > priority, self._clients)
        self._clients.insert(idx, (client, priority))

    def _remove_client(self, client):
        idx = index_if(lambda client_and_priority: client_and_priority[0] == client, self._clients)
        if idx != len(self._clients):
            del self._clients[idx]
            return True
        return None

    def _actual_owners(self):
        if self._clients:
            return [self._clients[-1][0]]
        return []

    @property
    def max_priority(self):
        if self._clients:
            return self._clients[-1][1]
        return DEFAULT_PRIORITY

    @property
    def stack_size(self):
        return len(self._clients)

    def get_owner(self):
        for owner in self._owners:
            return owner
        return None

    @property
    def clients(self):
        return list(map(first, self._clients))

    @property
    def owners(self):
        return self._owners

    def on_received(self, client):
        raise NotImplementedError("Override or pass callback")

    def on_lost(self, client):
        raise NotImplementedError("Override or pass callback")

    def release_stacked(self):
        clients = self.clients
        owners = self.owners
        for client in clients:
            if client not in owners:
                self.release(client)


class PrioritizedResource(StackingResource):
    def _actual_owners(self):
        max_priority = self.max_priority
        return [client for client, priority in self._clients if priority == max_priority]


class ClientWrapper(NamedTuple):
    wrap = partial(nop)
    unwrap = partial(nop)


class ProxyResource(Proxy):
    def __init__(self, proxied_resource=None, client_wrapper=ClientWrapper(), *a, **k):
        (super().__init__)(a, proxied_object=proxied_resource, **k)
        self._client_wrapper = client_wrapper

    def grab(self, client, *a, **k):
        (self.__getattr__("grab"))(self._client_wrapper.wrap(client), *a, **k)

    def release(self, client, *a, **k):
        (self.__getattr__("release"))(self._client_wrapper.wrap(client), *a, **k)

    @property
    def owner(self):
        return self._client_wrapper.unwrap(self.__getattr__("owner"))

    @property
    def owners(self):
        return list(map(self._client_wrapper.unwrap, self.__getattr__("owners")))



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/components/simple_device_navigation.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 2149 bytes
from typing import cast

import Live

from ...live import liveobj_changed, liveobj_valid
from ..display import Renderable
from . import Scrollable, ScrollComponent

NavDirection = Live.Application.Application.View.NavDirection


class SimpleDeviceNavigationComponent(ScrollComponent, Renderable, Scrollable):
    def __init__(self, name="Device_Navigation", *a, **k):
        (super().__init__)(a, name=name, scroll_skin_name="Device.Navigation", **k)
        self._previously_appointed_device = None

    def set_prev_button(self, button):
        self.scroll_up_button.set_control_element(button)

    def set_next_button(self, button):
        self.scroll_down_button.set_control_element(button)

    def can_scroll_up(self):
        return True

    def can_scroll_down(self):
        return True

    def scroll_up(self):
        self._scroll_device_chain(NavDirection.left)

    def scroll_down(self):
        self._scroll_device_chain(NavDirection.right)

    def _scroll_device_chain(self, direction):
        view = self.application.view
        if not (view.is_view_visible("Detail") and view.is_view_visible("Detail/DeviceChain")):
            view.show_view("Detail")
            view.show_view("Detail/DeviceChain")
        else:
            view.scroll_view(direction, "Detail/DeviceChain", False)
        self._tasks.add(self._notify_device_selection)

    def _notify_device_selection(self, _):
        device = self.song.appointed_device
        if liveobj_valid(device):
            if liveobj_changed(device, self._previously_appointed_device):
                self._previously_appointed_device = device
                self.notify(self.notifications.Device.select, cast(str, device.name))



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/display/view/view.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 4413 bytes
from typing import Callable, Generic

from ....base import const, depends
from ..type_decl import DISCONNECT_EVENT, INIT_EVENT, ContentType, Event, Render, State


class View(Generic[ContentType]):
    def __init__(
        self,
        render: Render[ContentType],
        render_condition: Callable[([State], bool)] = lambda _: True,
        content_condition: Callable[([ContentType], bool)] = lambda content: content is not None,
    ):
        self._render = render
        self._render_condition = render_condition
        self._content_condition = content_condition

    @depends(register_react_fn=(const(None)))
    def __new__(cls, *a, register_react_fn=None, **k):
        obj = (super().__new__)(cls, *a, **k)
        if hasattr(obj, "react"):
            register_react_fn(obj.react)
        return obj

    def __call__(self, state: State) -> ContentType:
        return self.render(state)

    def render(self, state: State) -> ContentType:
        return self._render(state)

    def render_condition(self, state: State) -> bool:
        return self._render_condition(state)

    def content_condition(self, content) -> bool:
        return self._content_condition(content)


class CompoundView(View[ContentType], Generic[ContentType]):
    class NoRender:
        pass

    def __init__(self, *views):
        super().__init__(self.compound_render)
        self._views = tuple(reversed(views))

    def compound_render(self, state: State) -> ContentType:
        result = self.NoRender()
        for view in self._views:
            if view.render_condition(state):
                content = view.render(state)
                if view.content_condition(content):
                    result = content

        return result


class DisconnectedView(View[ContentType], Generic[ContentType]):
    def __init__(
        self,
        render=const(None),
        render_condition=lambda state: not state.connected
        or hasattr(state, "identification")
        and not state.identification.is_identified,
        content_condition=lambda _: True,
    ):
        super().__init__(render, render_condition=render_condition, content_condition=content_condition)

    @staticmethod
    def react(state: State, event: Event):
        if event is INIT_EVENT:
            state.connected = True
        else:
            if event is DISCONNECT_EVENT:
                state.connected = False
        return state



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/controls/button.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 3019 bytes
from ...base import listenable_property
from ..display import Renderable
from . import ButtonControlBase, control_color


class ButtonControl(ButtonControlBase):
    class State(ButtonControlBase.State, Renderable):
        is_held = listenable_property.managed(False)
        color = control_color("DefaultButton.On")
        on_color = control_color(None)

        def __init__(self, color="DefaultButton.On", on_color=None, *a, **k):
            (super().__init__)(*a, **k)
            self.color = color
            self.on_color = on_color
            self._is_on = False

        @listenable_property
        def is_pressed(self):
            return self._is_pressed

        @property
        def is_on(self):
            return self._is_on

        @is_on.setter
        def is_on(self, is_on):
            if is_on != self._is_on:
                self._is_on = is_on
                self._send_current_color()

        def _send_button_color(self):
            if self.on_color is not None and self.is_on:
                self._control_element.set_light(self.on_color)
            else:
                if self.color is not None:
                    self._control_element.set_light(self.color)

        def _has_delayed_event(self):
            return True

        def _call_listener(self, listener_name, *a):
            (super()._call_listener)(listener_name, *a)
            if listener_name == "pressed":
                self.notify_is_pressed()
            else:
                if listener_name == "pressed_delayed":
                    self.is_held = True
                else:
                    if listener_name == "released":
                        self.is_held = False
                        self.notify_is_pressed()


class TouchControl(ButtonControl):
    class State(ButtonControl.State):
        def set_control_element(self, control_element):
            super().set_control_element(control_element)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_MxDCore/NotesAPIUtils.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 1246 bytes
MIDI_NOTE_ATTRS = (
    "note_id",
    "pitch",
    "start_time",
    "duration",
    "velocity",
    "mute",
    "probability",
    "velocity_deviation",
    "release_velocity",
)
REQUIRED_MIDI_NOTE_ATTRS = ("pitch", "start_time", "duration")
VALID_DUPLICATE_NOTES_BY_ID_PARAMETERS = ("note_ids", "destination_time", "transposition_amount")


def midi_note_to_dict(note, properties_to_return=None):
    def get_note_attr(attr_name):
        prop = getattr(note, attr_name)
        if attr_name == "mute":
            return int(prop)
        return prop

    note_dict = {}
    for attr in MIDI_NOTE_ATTRS:
        if properties_to_return is not None:
            if attr in properties_to_return:
                pass
        note_dict[attr] = get_note_attr(attr)

    return note_dict


def midi_notes_to_notes_dict(notes):
    return {"notes": [midi_note_to_dict(note) for note in notes]}


def verify_note_specification_requirements(note_specification):
    missing_keys = set(REQUIRED_MIDI_NOTE_ATTRS) - set(note_specification.keys())
    if len(missing_keys) > 0:
        raise RuntimeError("Missing required keys: ", ", ".join(missing_keys))



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/display/notifications/all.py
# Compiled at: 2023-12-18 13:26:15
# Size of source mod 2**32: 8711 bytes
from __future__ import annotations

from typing import NewType

from Live import ClipSlot

from ....base import pitch_index_to_string
from ....live import display_name, major_version
from .type_decl import Fn, Notification, _DefaultText, _TransformDefaultText
from .util import toggle_text_generator

SceneName = NewType("SceneName", str)
TrackName = NewType("TrackName", str)
ClipName = NewType("ClipName", str)
DeviceName = NewType("DeviceName", str)
DeviceBank = NewType("DeviceBank", str)
PadName = NewType("PadName", str)
ComponentName = NewType("ComponentName", str)
ModeName = NewType("ModeName", str)
Subdivision = NewType("Subdivision", str)
Resolution = NewType("Resolution", str)
Range = NewType("Range", str)
RangeName = NewType("RangeName", str)


class Notifications:
    identify = lambda: f"Live {major_version()}\nConnected"
    identify: Notification
    full_velocity = toggle_text_generator("Full Velocity\n{}")
    full_velocity: Notification[Fn[bool]]
    note_repeat = toggle_text_generator("Note Repeat\n{}")
    note_repeat: Notification[Fn[bool]]
    controlled_range = "{}\n{}".format
    controlled_range: Notification[Fn[RangeName, Range]]
    generic = "{}".format
    generic: Notification[Fn[str]]

    class Element:
        button_lock = lambda name, state: "{}\n{}".format(name.replace("_", " "), "locked" if state else "unlocked")
        button_lock: Notification[Fn[str, bool]]

    class Clipboard:
        clear = "Clipboard\ncleared"
        clear: Notification

    class UndoRedo:
        undo = "{}".format
        undo: Notification[Fn[str]]
        error_undo_no_action = "No Action To Undo"
        error_undo_no_action: Notification
        redo = "{}".format
        redo: Notification[Fn[str]]
        error_redo_no_action = "No Action To Redo"
        error_redo_no_action: Notification

    class Transport:
        metronome = toggle_text_generator("Metronome\n{}")
        metronome: Notification[Fn[bool]]
        midi_capture = (
            lambda tempo_set_by_capture, tempo: f"Captured\n{int(tempo)} BPM" if tempo_set_by_capture else "Captured"
        )
        midi_capture: Notification[Fn[bool, float]]
        loop = toggle_text_generator("Loop\n{}")
        loop: Notification[Fn[bool]]
        tap_tempo = lambda tempo: f"Tap Tempo\n{int(tempo)}"
        tap_tempo: Notification[Fn[float]]
        record_quantize = toggle_text_generator("Record Quantize\n{}")
        record_quantize: Notification[Fn[bool]]

    class Recording:
        new = "New Clip Slot\nselected"
        new: Notification

    class Automation:
        delete = "Automation\ndeleted"
        delete: Notification

    class Scene:
        select = "{}\nselected".format
        select: Notification[Fn[SceneName]]
        delete = "{}\ndeleted".format
        delete: Notification[Fn[SceneName]]

    class Track:
        lock = lambda name, state: "{}\n{}".format(name, "locked" if state else "unlocked")
        lock: Notification[Fn[TrackName, bool]]
        select = "{}".format
        select: Notification[Fn[TrackName]]
        delete = "{}\ndeleted".format
        delete: Notification[Fn[TrackName]]
        duplicate = "{}\nduplicated".format
        duplicate: Notification[Fn[TrackName]]
        mute = lambda name, state: "{}\n{}".format(name, "muted" if state else "unmuted")
        mute: Notification[Fn[TrackName, bool]]

    class Clip:
        select = "{}\nselected".format
        select: Notification[Fn[ClipName]]
        delete = "{}\ndeleted".format
        delete: Notification[Fn[ClipName]]
        duplicate = "{}\nduplicated".format
        duplicate: Notification[Fn[ClipName]]
        error_delete_empty_slot = "Clip Slot\nalready empty"
        error_delete_empty_slot: Notification
        quantize = "{} {}\nquantized".format
        quantize: Notification[Fn[ClipName, Subdivision]]
        error_quantize_invalid_resolution = "Cannot quantize to {}".format
        error_quantize_invalid_resolution: Notification[Fn[Resolution]]
        double_loop = "Loop\ndoubled"
        double_loop: Notification

        class CopyPaste:
            error_copy_from_group_slot = "Cannot copy from Group Slot"
            error_copy_from_group_slot: Notification
            error_copy_from_empty_slot = "Cannot copy from empty Slot"
            error_copy_from_empty_slot: Notification
            error_copy_recording_clip = "Cannot copy recording Clip"
            error_copy_recording_clip: Notification
            copy = lambda slot: f"{display_name(slot)}\ncopied"
            copy: Notification[Fn[ClipSlot]]
            error_paste_to_group_slot = "Cannot paste into Group Slot"
            error_paste_to_group_slot: Notification
            error_paste_audio_to_midi = "Cannot paste an audio Clip into a MIDI Track"
            error_paste_audio_to_midi: Notification
            error_paste_midi_to_audio = "Cannot paste a MIDI Clip into an audio Track"
            error_paste_midi_to_audio: Notification
            paste = lambda slot: f"{display_name(slot)}\npasted"
            paste: Notification[Fn[ClipSlot]]

    class Device:
        lock = lambda name, state: "{}\n{}".format(name, "locked" if state else "unlocked")
        lock: Notification[Fn[DeviceName, bool]]
        select = "{}".format
        select: Notification[Fn[DeviceName]]
        bank = "{}".format
        bank: Notification[Fn[DeviceBank]]

    class DrumGroup:
        class Pad:
            select = "{}\nselected".format
            select: Notification[Fn[PadName]]
            delete = "{}\ndeleted".format
            delete: Notification[Fn[PadName]]
            mute = lambda name, state: "{}\n{}".format(name, "muted" if state else "unmuted")
            mute: Notification[Fn[PadName]]
            delete_notes = "{}\nnotes deleted".format
            delete_notes: Notification[Fn[PadName]]

            class CopyPaste:
                error_copy_from_empty_pad = "Cannot copy from empty Pad"
                error_copy_from_empty_pad: Notification
                copy = "{}\ncopied".format
                copy: Notification[Fn[PadName]]
                error_paste_to_source_pad = "Cannot paste to source Pad"
                error_paste_to_source_pad: Notification
                paste = "{}\npasted".format
                paste: Notification[Fn[PadName]]

        class Page:
            up = "Page up"
            up: Notification
            down = "Page down"
            down: Notification

        class Scroll:
            up = "Scroll up"
            up: Notification
            down = "Scroll down"
            down: Notification

    class Simpler:
        class Slice:
            select = "Slice {}\nselected".format
            select: Notification[Fn[int]]
            delete = "Slice {}\ndeleted".format
            delete: Notification[Fn[int]]
            delete_notes = "Slice {}\nnotes deleted".format
            delete_notes: Notification[Fn[int]]

        class Page:
            up = "Page up"
            up: Notification
            down = "Page down"
            down: Notification

        class Scroll:
            up = "Scroll up"
            up: Notification
            down = "Scroll down"
            down: Notification

    class Notes:
        delete = "Notes\ndeleted"
        delete: Notification
        error_no_notes_to_delete = "No notes\nto delete"
        error_no_notes_to_delete: Notification
        nudge = "Notes\nnudged"
        nudge: Notification

        class Pitch:
            select = lambda index: f"Pitch {pitch_index_to_string(index)}\nselected"
            select: Notification[Fn[int]]
            delete = lambda index: f"{pitch_index_to_string(index)}\nnotes deleted"
            delete: Notification[Fn[int]]

        class Octave:
            up = "Octave up"
            up: Notification
            down = "Octave down"
            down: Notification

        class ScaleDegree:
            up = "Scale degree\nup"
            up: Notification
            down = "Scale degree\ndown"
            down: Notification

    class Modes:
        select = None
        select: Notification[Fn[ComponentName, ModeName]]

    class DefaultText(_DefaultText):
        pass

    class TransformDefaultText(_TransformDefaultText):
        pass



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/components/background.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 4615 bytes
from functools import partial

from .. import Component
from ..controls import ButtonControl, InputControl, RadioButtonGroup


class NopControl(InputControl):
    class State(InputControl.State):
        def _event_listener_required(self):
            return True


class BackgroundComponent(Component):
    def __init__(self, name="Background", control_type=None, *a, **k):
        (super().__init__)(a, name=name, **k)
        self._control_type = control_type or NopControl

    def _setup_control_state(self, name, control_state):
        pass

    def __getattr__(self, name):
        if name.startswith("set_"):
            return partial(self._set_element_for_control, name[4:])
        raise AttributeError(name)

    def _set_element_for_control(self, name, element):
        if element:
            element.reset()
        if name in self.__dict__:
            self.__dict__[name].set_control_element(element)
        else:
            control = self._control_type()
            control_state = control._get_state(self)
            setattr(self, name, control_state)
            self._setup_control_state(name, control_state)
            control_state.set_control_element(element)


class TranslatingBackgroundComponent(BackgroundComponent):
    channel_selection_buttons = RadioButtonGroup(
        unchecked_color="Translation.Channel.NotSelected",
        checked_color="Translation.Channel.Selected",
    )

    def __init__(self, name="Translating_Background", translation_channel=0, *a, **k):
        (super().__init__)(a, name=name, **k)
        self._base_translation_channel = translation_channel

    def set_channel_selection_buttons(self, buttons):
        self.channel_selection_buttons.set_control_element(buttons)
        if buttons:
            if self.channel_selection_buttons.checked_index < 0:
                self.channel_selection_buttons.checked_index = 0

    @channel_selection_buttons.checked
    def channel_selection_buttons(self, button):
        for control_state in self._control_states.values():
            control_state.channel = self._base_translation_channel + button.index

    def _setup_control_state(self, _, control_state):
        control_state.channel = self._base_translation_channel


class ModifierBackgroundComponent(BackgroundComponent):
    def __init__(self, name="Modifier_Background", *a, **k):
        (super().__init__)(a, name=name, control_type=ButtonControl, **k)

    def _setup_control_state(self, name, control_state):
        base_name = self.name.title().replace("_", "")
        control_name = name.title().replace("_", "")
        control_state.color = f"{base_name}.{control_name}"
        control_state.pressed_color = f"{base_name}.{control_name}Pressed"



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/device_chain_utils.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 1446 bytes
from functools import partial
from itertools import chain

import Live

from ..base import find_if, liveobj_valid


def find_instrument_devices(track_or_chain):
    if liveobj_valid(track_or_chain):
        instrument = find_if(lambda d: d.type == Live.Device.DeviceType.instrument, track_or_chain.devices)
        if liveobj_valid(instrument):
            if not instrument.can_have_drum_pads:
                if instrument.can_have_chains:
                    return chain([instrument], *map(find_instrument_devices, instrument.chains))
            return [instrument]
    return []


def find_instrument_meeting_requirement(requirement, track_or_chain):
    if liveobj_valid(track_or_chain):
        instrument = find_if(lambda d: d.type == Live.Device.DeviceType.instrument, track_or_chain.devices)
        if liveobj_valid(instrument):
            if requirement(instrument):
                return instrument
            if instrument.can_have_chains:
                recursive_call = partial(find_instrument_meeting_requirement, requirement)
                return find_if(bool, map(recursive_call, instrument.chains))
            return None
        return None
    return None



from __future__ import annotations

from typing import Any, Callable

from Live import *
from Live.Application import *
from Live.Base import *
from Live.Browser import *
from Live.CcControlDevice import *
from Live.Chain import *
from Live.ChainMixerDevice import *
from Live.Clip import *
from Live.ClipSlot import *
from Live.CompressorDevice import *
from Live.Conversions import *
from Live.Device import *
from Live.DeviceIO import *
from Live.DeviceParameter import *
from Live.DriftDevice import *
from Live.DrumChain import *
from Live.DrumPad import *
from Live.Eq8Device import *
from Live.Groove import *
from Live.GroovePool import *
from Live.HybridReverbDevice import *
from Live.Listener import *
from Live.LomObject import *
from Live.MaxDevice import *
from Live.MeldDevice import *
from Live.MidiMap import *
from Live.MixerDevice import *
from Live.PluginDevice import *
from Live.RoarDevice import *
from Live.Sample import *
from Live.Scene import *
from Live.ShifterDevice import *
from Live.SimplerDevice import *
from Live.Song import *
from Live.SpectralResonatorDevice import *
from Live.Track import *
from Live.WavetableDevice import *


class RackDevice:
    """
    This class represents a Rack device.
    """

    class View:
        """
        Representing the view aspects of a rack device.
        """

        def add_drum_pads_scroll_position_listener(self, listener: Callable) -> None:
            """
            add_drum_pads_scroll_position_listener( (View)arg1, (object)arg2) -> None :
                Add a listener function or method, which will be called as soon as the
                property "drum_pads_scroll_position" has changed.

                C++ signature :
                    void add_drum_pads_scroll_position_listener(TRackDevicePyViewData,boost::python::api::object)
            """

        def add_is_collapsed_listener(self, listener: Callable) -> None:
            """
            add_is_collapsed_listener( (View)arg1, (object)arg2) -> None :
                Add a listener function or method, which will be called as soon as the
                property "is_collapsed" has changed.

                C++ signature :
                    void add_is_collapsed_listener(TPyViewData<ADevice>,boost::python::api::object)
            """

        def add_is_showing_chain_devices_listener(self, listener: Callable) -> None:
            """
            add_is_showing_chain_devices_listener( (View)arg1, (object)arg2) -> None :
                Add a listener function or method, which will be called as soon as the
                property "is_showing_chain_devices" has changed.

                C++ signature :
                    void add_is_showing_chain_devices_listener(TRackDevicePyViewData,boost::python::api::object)
            """

        def add_selected_chain_listener(self, listener: Callable) -> None:
            """
            add_selected_chain_listener( (View)arg1, (object)arg2) -> None :
                Add a listener function or method, which will be called as soon as the
                property "selected_chain" has changed.

                C++ signature :
                    void add_selected_chain_listener(TRackDevicePyViewData,boost::python::api::object)
            """

        def add_selected_drum_pad_listener(self, listener: Callable) -> None:
            """
            add_selected_drum_pad_listener( (View)arg1, (object)arg2) -> None :
                Add a listener function or method, which will be called as soon as the
                property "selected_drum_pad" has changed.

                C++ signature :
                    void add_selected_drum_pad_listener(TRackDevicePyViewData,boost::python::api::object)
            """

        def drum_pads_scroll_position_has_listener(self, listener: Callable) -> bool:
            """
            drum_pads_scroll_position_has_listener( (View)arg1, (object)arg2) -> bool :
                Returns true, if the given listener function or method is connected
                to the property "drum_pads_scroll_position".

                C++ signature :
                    bool drum_pads_scroll_position_has_listener(TRackDevicePyViewData,boost::python::api::object)
            """

        def is_collapsed_has_listener(self, listener: Callable) -> bool:
            """
            is_collapsed_has_listener( (View)arg1, (object)arg2) -> bool :
                Returns true, if the given listener function or method is connected
                to the property "is_collapsed".

                C++ signature :
                    bool is_collapsed_has_listener(TPyViewData<ADevice>,boost::python::api::object)
            """

        def is_showing_chain_devices_has_listener(self, listener: Callable) -> bool:
            """
            is_showing_chain_devices_has_listener( (View)arg1, (object)arg2) -> bool :
                Returns true, if the given listener function or method is connected
                to the property "is_showing_chain_devices".

                C++ signature :
                    bool is_showing_chain_devices_has_listener(TRackDevicePyViewData,boost::python::api::object)
            """

        def remove_drum_pads_scroll_position_listener(self, listener: Callable) -> None:
            """
            remove_drum_pads_scroll_position_listener( (View)arg1, (object)arg2) -> None :
                Remove a previously set listener function or method from
                property "drum_pads_scroll_position".

                C++ signature :
                    void remove_drum_pads_scroll_position_listener(TRackDevicePyViewData,boost::python::api::object)
            """

        def remove_is_collapsed_listener(self, listener: Callable) -> None:
            """
            remove_is_collapsed_listener( (View)arg1, (object)arg2) -> None :
                Remove a previously set listener function or method from
                property "is_collapsed".

                C++ signature :
                    void remove_is_collapsed_listener(TPyViewData<ADevice>,boost::python::api::object)
            """

        def remove_is_showing_chain_devices_listener(self, listener: Callable) -> None:
            """
            remove_is_showing_chain_devices_listener( (View)arg1, (object)arg2) -> None :
                Remove a previously set listener function or method from
                property "is_showing_chain_devices".

                C++ signature :
                    void remove_is_showing_chain_devices_listener(TRackDevicePyViewData,boost::python::api::object)
            """

        def remove_selected_chain_listener(self, listener: Callable) -> None:
            """
            remove_selected_chain_listener( (View)arg1, (object)arg2) -> None :
                Remove a previously set listener function or method from
                property "selected_chain".

                C++ signature :
                    void remove_selected_chain_listener(TRackDevicePyViewData,boost::python::api::object)
            """

        def remove_selected_drum_pad_listener(self, listener: Callable) -> None:
            """
            remove_selected_drum_pad_listener( (View)arg1, (object)arg2) -> None :
                Remove a previously set listener function or method from
                property "selected_drum_pad".

                C++ signature :
                    void remove_selected_drum_pad_listener(TRackDevicePyViewData,boost::python::api::object)
            """

        def selected_chain_has_listener(self, listener: Callable) -> bool:
            """
            selected_chain_has_listener( (View)arg1, (object)arg2) -> bool :
                Returns true, if the given listener function or method is connected
                to the property "selected_chain".

                C++ signature :
                    bool selected_chain_has_listener(TRackDevicePyViewData,boost::python::api::object)
            """

        def selected_drum_pad_has_listener(self, listener: Callable) -> bool:
            """
            selected_drum_pad_has_listener( (View)arg1, (object)arg2) -> bool :
                Returns true, if the given listener function or method is connected
                to the property "selected_drum_pad".

                C++ signature :
                    bool selected_drum_pad_has_listener(TRackDevicePyViewData,boost::python::api::object)
            """

        @property
        def canonical_parent(self) -> Any:
            """
            Get the canonical parent of the View.
            """

        @property
        def drum_pads_scroll_position(self) -> Any:
            """
            Access to the index of the lowest visible row of pads. Throws an exception if can_have_drum_pads is false.
            """

        @property
        def is_collapsed(self) -> bool:
            """
            Get/Set/Listen if the device is shown collapsed in the device chain.
            """

        @is_collapsed.setter
        def is_collapsed(self, value: bool) -> None:
            pass

        @property
        def is_showing_chain_devices(self) -> bool:
            """
            Return whether the devices in the currently selected chain are visible. Throws an exception if can_have_chains is false.
            """

        @property
        def selected_chain(self) -> Any:
            """
            Return access to the currently selected chain.
            """

        @property
        def selected_drum_pad(self) -> Any:
            """
            Return access to the currently selected drum pad. Throws an exception if can_have_drum_pads is false.
            """

    def add_chains_listener(self, listener: Callable) -> None:
        """
        add_chains_listener( (RackDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "chains" has changed.

            C++ signature :
                void add_chains_listener(TRackDevicePyHandle,boost::python::api::object)
        """

    def add_drum_pads_listener(self, listener: Callable) -> None:
        """
        add_drum_pads_listener( (RackDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "drum_pads" has changed.

            C++ signature :
                void add_drum_pads_listener(TRackDevicePyHandle,boost::python::api::object)
        """

    def add_has_drum_pads_listener(self, listener: Callable) -> None:
        """
        add_has_drum_pads_listener( (RackDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "has_drum_pads" has changed.

            C++ signature :
                void add_has_drum_pads_listener(TRackDevicePyHandle,boost::python::api::object)
        """

    def add_has_macro_mappings_listener(self, listener: Callable) -> None:
        """
        add_has_macro_mappings_listener( (RackDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "has_macro_mappings" has changed.

            C++ signature :
                void add_has_macro_mappings_listener(TRackDevicePyHandle,boost::python::api::object)
        """

    def add_is_active_listener(self, listener: Callable) -> None:
        """
        add_is_active_listener( (Device)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "is_active" has changed.

            C++ signature :
                void add_is_active_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def add_is_showing_chains_listener(self, listener: Callable) -> None:
        """
        add_is_showing_chains_listener( (RackDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "is_showing_chains" has changed.

            C++ signature :
                void add_is_showing_chains_listener(TRackDevicePyHandle,boost::python::api::object)
        """

    def add_latency_in_ms_listener(self, listener: Callable) -> None:
        """
        add_latency_in_ms_listener( (Device)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "latency_in_ms" has changed.

            C++ signature :
                void add_latency_in_ms_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def add_latency_in_samples_listener(self, listener: Callable) -> None:
        """
        add_latency_in_samples_listener( (Device)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "latency_in_samples" has changed.

            C++ signature :
                void add_latency_in_samples_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def add_macro(self) -> None:
        """
        add_macro( (RackDevice)arg1) -> None :
            Increases the number of visible macro controls in the rack. Throws an exception if the maximum number of macro controls is reached.

            C++ signature :
                void add_macro(TRackDevicePyHandle)
        """

    def add_macros_mapped_listener(self, listener: Callable) -> None:
        """
        add_macros_mapped_listener( (RackDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "macros_mapped" has changed.

            C++ signature :
                void add_macros_mapped_listener(TRackDevicePyHandle,boost::python::api::object)
        """

    def add_name_listener(self, listener: Callable) -> None:
        """
        add_name_listener( (Device)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "name" has changed.

            C++ signature :
                void add_name_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def add_parameters_listener(self, listener: Callable) -> None:
        """
        add_parameters_listener( (Device)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "parameters" has changed.

            C++ signature :
                void add_parameters_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def add_return_chains_listener(self, listener: Callable) -> None:
        """
        add_return_chains_listener( (RackDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "return_chains" has changed.

            C++ signature :
                void add_return_chains_listener(TRackDevicePyHandle,boost::python::api::object)
        """

    def add_variation_count_listener(self, listener: Callable) -> None:
        """
        add_variation_count_listener( (RackDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "variation_count" has changed.

            C++ signature :
                void add_variation_count_listener(TRackDevicePyHandle,boost::python::api::object)
        """

    def add_visible_drum_pads_listener(self, listener: Callable) -> None:
        """
        add_visible_drum_pads_listener( (RackDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "visible_drum_pads" has changed.

            C++ signature :
                void add_visible_drum_pads_listener(TRackDevicePyHandle,boost::python::api::object)
        """

    def add_visible_macro_count_listener(self, listener: Callable) -> None:
        """
        add_visible_macro_count_listener( (RackDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "visible_macro_count" has changed.

            C++ signature :
                void add_visible_macro_count_listener(TRackDevicePyHandle,boost::python::api::object)
        """

    def chains_has_listener(self, listener: Callable) -> bool:
        """
        chains_has_listener( (RackDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "chains".

            C++ signature :
                bool chains_has_listener(TRackDevicePyHandle,boost::python::api::object)
        """

    def copy_pad(self, arg2: int, arg3: int) -> None:
        """
        copy_pad( (RackDevice)arg1, (int)arg2, (int)arg3) -> None :
            Copies all contents of a drum pad from a source pad into a destination pad. copy_pad(source_index, destination_index) where source_index and destination_index correspond to the note number/index of the drum pad in a drum rack. Throws an exception when the source pad is empty, or when the source or destination indices are not between 0 - 127.

            C++ signature :
                void copy_pad(TRackDevicePyHandle,int,int)
        """

    def delete_selected_variation(self) -> None:
        """
        delete_selected_variation( (Device)arg1) -> None :
            Deletes the currently selected macro variation.Does nothing if there is no selected variation.

            C++ signature :
                void delete_selected_variation(TPyHandle<ADevice>)
        """

    def drum_pads_has_listener(self, listener: Callable) -> bool:
        """
        drum_pads_has_listener( (RackDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "drum_pads".

            C++ signature :
                bool drum_pads_has_listener(TRackDevicePyHandle,boost::python::api::object)
        """

    def has_drum_pads_has_listener(self, listener: Callable) -> bool:
        """
        has_drum_pads_has_listener( (RackDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "has_drum_pads".

            C++ signature :
                bool has_drum_pads_has_listener(TRackDevicePyHandle,boost::python::api::object)
        """

    def has_macro_mappings_has_listener(self, listener: Callable) -> bool:
        """
        has_macro_mappings_has_listener( (RackDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "has_macro_mappings".

            C++ signature :
                bool has_macro_mappings_has_listener(TRackDevicePyHandle,boost::python::api::object)
        """

    def is_active_has_listener(self, listener: Callable) -> bool:
        """
        is_active_has_listener( (Device)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "is_active".

            C++ signature :
                bool is_active_has_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def is_showing_chains_has_listener(self, listener: Callable) -> bool:
        """
        is_showing_chains_has_listener( (RackDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "is_showing_chains".

            C++ signature :
                bool is_showing_chains_has_listener(TRackDevicePyHandle,boost::python::api::object)
        """

    def latency_in_ms_has_listener(self, listener: Callable) -> bool:
        """
        latency_in_ms_has_listener( (Device)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "latency_in_ms".

            C++ signature :
                bool latency_in_ms_has_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def latency_in_samples_has_listener(self, listener: Callable) -> bool:
        """
        latency_in_samples_has_listener( (Device)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "latency_in_samples".

            C++ signature :
                bool latency_in_samples_has_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def macros_mapped_has_listener(self, listener: Callable) -> bool:
        """
        macros_mapped_has_listener( (RackDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "macros_mapped".

            C++ signature :
                bool macros_mapped_has_listener(TRackDevicePyHandle,boost::python::api::object)
        """

    def name_has_listener(self, listener: Callable) -> bool:
        """
        name_has_listener( (Device)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "name".

            C++ signature :
                bool name_has_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def parameters_has_listener(self, listener: Callable) -> bool:
        """
        parameters_has_listener( (Device)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "parameters".

            C++ signature :
                bool parameters_has_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def randomize_macros(self) -> None:
        """
        randomize_macros( (RackDevice)arg1) -> None :
            Randomizes the values for all macro controls not excluded from randomization.

            C++ signature :
                void randomize_macros(TRackDevicePyHandle)
        """

    def recall_last_used_variation(self) -> None:
        """
        recall_last_used_variation( (Device)arg1) -> None :
            Recalls the macro variation that was recalled most recently.Does nothing if no variation has been recalled yet.

            C++ signature :
                void recall_last_used_variation(TPyHandle<ADevice>)
        """

    def recall_selected_variation(self) -> None:
        """
        recall_selected_variation( (Device)arg1) -> None :
            Recalls the currently selected macro variation.Does nothing if there are no variations.

            C++ signature :
                void recall_selected_variation(TPyHandle<ADevice>)
        """

    def remove_chains_listener(self, listener: Callable) -> None:
        """
        remove_chains_listener( (RackDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "chains".

            C++ signature :
                void remove_chains_listener(TRackDevicePyHandle,boost::python::api::object)
        """

    def remove_drum_pads_listener(self, listener: Callable) -> None:
        """
        remove_drum_pads_listener( (RackDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "drum_pads".

            C++ signature :
                void remove_drum_pads_listener(TRackDevicePyHandle,boost::python::api::object)
        """

    def remove_has_drum_pads_listener(self, listener: Callable) -> None:
        """
        remove_has_drum_pads_listener( (RackDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "has_drum_pads".

            C++ signature :
                void remove_has_drum_pads_listener(TRackDevicePyHandle,boost::python::api::object)
        """

    def remove_has_macro_mappings_listener(self, listener: Callable) -> None:
        """
        remove_has_macro_mappings_listener( (RackDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "has_macro_mappings".

            C++ signature :
                void remove_has_macro_mappings_listener(TRackDevicePyHandle,boost::python::api::object)
        """

    def remove_is_active_listener(self, listener: Callable) -> None:
        """
        remove_is_active_listener( (Device)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "is_active".

            C++ signature :
                void remove_is_active_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def remove_is_showing_chains_listener(self, listener: Callable) -> None:
        """
        remove_is_showing_chains_listener( (RackDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "is_showing_chains".

            C++ signature :
                void remove_is_showing_chains_listener(TRackDevicePyHandle,boost::python::api::object)
        """

    def remove_latency_in_ms_listener(self, listener: Callable) -> None:
        """
        remove_latency_in_ms_listener( (Device)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "latency_in_ms".

            C++ signature :
                void remove_latency_in_ms_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def remove_latency_in_samples_listener(self, listener: Callable) -> None:
        """
        remove_latency_in_samples_listener( (Device)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "latency_in_samples".

            C++ signature :
                void remove_latency_in_samples_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def remove_macro(self) -> None:
        """
        remove_macro( (RackDevice)arg1) -> None :
            Decreases the number of visible macro controls in the rack. Throws an exception if the minimum number of macro controls is reached.

            C++ signature :
                void remove_macro(TRackDevicePyHandle)
        """

    def remove_macros_mapped_listener(self, listener: Callable) -> None:
        """
        remove_macros_mapped_listener( (RackDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "macros_mapped".

            C++ signature :
                void remove_macros_mapped_listener(TRackDevicePyHandle,boost::python::api::object)
        """

    def remove_name_listener(self, listener: Callable) -> None:
        """
        remove_name_listener( (Device)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "name".

            C++ signature :
                void remove_name_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def remove_parameters_listener(self, listener: Callable) -> None:
        """
        remove_parameters_listener( (Device)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "parameters".

            C++ signature :
                void remove_parameters_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def remove_return_chains_listener(self, listener: Callable) -> None:
        """
        remove_return_chains_listener( (RackDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "return_chains".

            C++ signature :
                void remove_return_chains_listener(TRackDevicePyHandle,boost::python::api::object)
        """

    def remove_variation_count_listener(self, listener: Callable) -> None:
        """
        remove_variation_count_listener( (RackDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "variation_count".

            C++ signature :
                void remove_variation_count_listener(TRackDevicePyHandle,boost::python::api::object)
        """

    def remove_visible_drum_pads_listener(self, listener: Callable) -> None:
        """
        remove_visible_drum_pads_listener( (RackDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "visible_drum_pads".

            C++ signature :
                void remove_visible_drum_pads_listener(TRackDevicePyHandle,boost::python::api::object)
        """

    def remove_visible_macro_count_listener(self, listener: Callable) -> None:
        """
        remove_visible_macro_count_listener( (RackDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "visible_macro_count".

            C++ signature :
                void remove_visible_macro_count_listener(TRackDevicePyHandle,boost::python::api::object)
        """

    def return_chains_has_listener(self, listener: Callable) -> bool:
        """
        return_chains_has_listener( (RackDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "return_chains".

            C++ signature :
                bool return_chains_has_listener(TRackDevicePyHandle,boost::python::api::object)
        """

    def store_chosen_bank(self, arg2: int, arg3: int) -> None:
        """
        store_chosen_bank( (Device)arg1, (int)arg2, (int)arg3) -> None :
            Set the selected bank in the device for persistency.

            C++ signature :
                void store_chosen_bank(TPyHandle<ADevice>,int,int)
        """

    def store_variation(self) -> None:
        """
        store_variation( (Device)arg1) -> None :
            Stores a new variation of the values of all currently mapped macros

            C++ signature :
                void store_variation(TPyHandle<ADevice>)
        """

    def variation_count_has_listener(self, listener: Callable) -> bool:
        """
        variation_count_has_listener( (RackDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "variation_count".

            C++ signature :
                bool variation_count_has_listener(TRackDevicePyHandle,boost::python::api::object)
        """

    def visible_drum_pads_has_listener(self, listener: Callable) -> bool:
        """
        visible_drum_pads_has_listener( (RackDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "visible_drum_pads".

            C++ signature :
                bool visible_drum_pads_has_listener(TRackDevicePyHandle,boost::python::api::object)
        """

    def visible_macro_count_has_listener(self, listener: Callable) -> bool:
        """
        visible_macro_count_has_listener( (RackDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "visible_macro_count".

            C++ signature :
                bool visible_macro_count_has_listener(TRackDevicePyHandle,boost::python::api::object)
        """

    @property
    def can_have_chains(self) -> bool:
        """
        Returns true if the device is a rack.
        """

    @property
    def can_have_drum_pads(self) -> bool:
        """
        Returns true if the device is a drum rack.
        """

    @property
    def can_show_chains(self) -> bool:
        """
        return True, if this Rack contains a rack instrument device that is capable of showing its chains in session view.
        """

    @property
    def canonical_parent(self) -> Any:
        """
        Get the canonical parent of the Device.
        """

    @property
    def chain_selector(self) -> Any:
        """
        Const access to the chain selector parameter.
        """

    @property
    def chains(self) -> Any:
        """
        Return const access to the list of chains in this device. Throws an exception if can_have_chains is false.
        """

    @property
    def class_display_name(self) -> str:
        """
        Return const access to the name of the device's class name as displayed in Live's browser and device chain
        """

    @property
    def class_name(self) -> str:
        """
        Return const access to the name of the device's class.
        """

    @property
    def drum_pads(self) -> Any:
        """
        Return const access to the list of drum pads in this device. Throws an exception if can_have_drum_pads is false.
        """

    @property
    def has_drum_pads(self) -> bool:
        """
        Returns true if the device is a drum rack which has drum pads. Throws an exception if can_have_drum_pads is false.
        """

    @property
    def has_macro_mappings(self) -> bool:
        """
        Returns true if any of the rack's macros are mapped to a parameter.
        """

    @property
    def is_active(self) -> bool:
        """
        Return const access to whether this device is active. This will be false bothwhen the device is off and when it's inside a rack device which is off.
        """

    @property
    def is_showing_chains(self) -> bool:
        """
        Returns True, if it is showing chains.
        """

    @property
    def latency_in_ms(self) -> Any:
        """
        Returns the latency of the device in ms.
        """

    @property
    def latency_in_samples(self) -> Any:
        """
        Returns the latency of the device in samples.
        """

    @property
    def macros_mapped(self) -> Any:
        """
        A list of booleans, one for each macro parameter, which is True iffthat macro is mapped to something
        """

    @property
    def name(self) -> str:
        """
        Return access to the name of the device.
        """

    @property
    def parameters(self) -> list[DeviceParameter]:
        """
        Const access to the list of available automatable parameters for this device.
        """

    @property
    def return_chains(self) -> Any:
        """
        Return const access to the list of return chains in this device. Throws an exception if can_have_chains is false.
        """

    @property
    def selected_variation_index(self) -> Any:
        """
        Access to the index of the currently selected macro variation.Throws an exception if the index is out of range.
        """

    @property
    def type(self) -> Any:
        """
        Return the type of the device.
        """

    @property
    def variation_count(self) -> Any:
        """
        Access to the number of macro variations currently stored.
        """

    @property
    def view(self) -> View:
        """
        Representing the view aspects of a device.
        """

    @property
    def visible_drum_pads(self) -> Any:
        """
        Return const access to the list of visible drum pads in this device. Throws an exception if can_have_drum_pads is false.
        """

    @property
    def visible_macro_count(self) -> Any:
        """
        Access to the number of macros that are currently visible.
        """



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_MxDCore/MxDUtils.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 9496 bytes
import logging

from ableton.v2.base import old_hasattr
from future.utils import string_types

logger = logging.getLogger(__name__)


class TupleWrapper:
    _tuple_wrapper_registry = {}

    def forget_tuple_wrapper_instances():
        TupleWrapper._tuple_wrapper_registry = {}

    forget_tuple_wrapper_instances = staticmethod(forget_tuple_wrapper_instances)

    def get_tuple_wrapper(parent, attribute, element_filter=None, element_transform=None):
        if (parent, attribute) not in TupleWrapper._tuple_wrapper_registry:
            TupleWrapper._tuple_wrapper_registry[(parent, attribute)] = TupleWrapper(
                parent,
                attribute,
                element_filter,
                element_transform,
            )
        return TupleWrapper._tuple_wrapper_registry[(parent, attribute)]

    get_tuple_wrapper = staticmethod(get_tuple_wrapper)

    def __init__(self, parent, attribute, element_filter=None, element_transform=None):
        self._parent = parent
        self._attribute = attribute
        self._element_filter = element_filter
        self._element_transform = element_transform

    def get_list(self):
        result = ()
        parent = self._parent
        if isinstance(parent, dict):
            if self._attribute in list(parent.keys()):
                result = parent[self._attribute]
        else:
            if old_hasattr(parent, self._attribute):
                result = getattr(parent, self._attribute)
        filtered_result = [e if self._element_filter(e) else None for e in result] if self._element_filter else result
        if self._element_transform:
            return list(map(self._element_transform, filtered_result))
        return filtered_result


STATE_NEUTRAL = "neutral"
STATE_QUOTED_STR = "quoted"
STATE_UNQUOTED_STR = "unquoted"
STATE_PENDING_NR = "number"
STATE_PENDING_FLOAT = "float"
QUOTE_ENTITY = "&quot;"
QUOTE_SIMPLE = '"'


class StringHandler:
    def prepare_incoming(string):
        return string.replace(QUOTE_ENTITY, QUOTE_SIMPLE)

    prepare_incoming = staticmethod(prepare_incoming)

    def prepare_outgoing(string):
        result = string.replace(QUOTE_SIMPLE, QUOTE_ENTITY)
        if result.find(" ") >= 0:
            result = QUOTE_SIMPLE + result + QUOTE_SIMPLE
        return result

    prepare_outgoing = staticmethod(prepare_outgoing)

    def parse(string, id_callback):
        return StringHandler(id_callback).parse_string(string)

    parse = staticmethod(parse)

    def __init__(self, id_callback):
        self._state = STATE_NEUTRAL
        self._sub_string = ""
        self._open_quote_index = -1
        self._id_callback = id_callback

    def parse_string(self, string):
        self._arguments = []
        self._sub_string = ""
        self._state = STATE_NEUTRAL
        self._open_quote_index = -1
        for index in range(len(string)):
            char = string[index]
            handle_selector = "_" + str(self._state) + "_handle_char"
            if old_hasattr(self, handle_selector):
                getattr(self, handle_selector)(char, index)
            else:
                logger.info("Unknown state " + str(self._state))

        finalize_selector = "_finalize_" + str(self._state)
        if len(self._sub_string) > 0:
            if old_hasattr(self, finalize_selector):
                getattr(self, finalize_selector)()
        return self._arguments

    def _neutral_handle_char(self, char, index):
        if char == '"':
            self._open_quote_index = index
            self._state = STATE_QUOTED_STR
        else:
            if char != " ":
                self._sub_string += char
                if str(char).isdigit():
                    self._state = STATE_PENDING_NR
                else:
                    self._state = STATE_UNQUOTED_STR

    def _number_handle_char(self, char, index):
        if char == " ":
            if len(self._sub_string) > 0:
                self._finalize_number()
            else:
                self._state = STATE_NEUTRAL
        else:
            if char == ".":
                self._state = STATE_PENDING_FLOAT
            else:
                if not str(char).isdigit():
                    self._state = STATE_UNQUOTED_STR
            self._sub_string += char

    def _float_handle_char(self, char, index):
        if char == " ":
            self._add_argument(float(self._sub_string))
        else:
            if char in (".", "e", "E"):
                if char in self._sub_string:
                    self._state = STATE_UNQUOTED_STR
            else:
                if not str(char).isdigit():
                    self._state = STATE_UNQUOTED_STR
            self._sub_string += char

    def _unquoted_handle_char(self, char, index):
        if char == " ":
            self._add_argument(self._sub_string)
        else:
            if str(char).isdigit():
                if self._sub_string == "-":
                    self._state = STATE_PENDING_NR
                else:
                    if self._sub_string in (".", "-."):
                        self._state = STATE_PENDING_FLOAT
            self._sub_string += char

    def _quoted_handle_char(self, char, index):
        if char == '"':
            self._open_quote_index = -1
            self._add_argument(self._sub_string)
        else:
            self._sub_string += char

    def _finalize_quoted(self):
        raise RuntimeError("no match for quote at index %d found" % self._open_quote_index)

    def _finalize_unquoted(self):
        self._add_argument(self._sub_string)

    def _finalize_float(self):
        self._add_argument(float(self._sub_string))

    def _finalize_number(self):
        argument = int(self._sub_string)
        if str(self._arguments[-1]) == "id":
            self._arguments.pop()
            try:
                argument = self._id_callback(argument)
            except KeyError:
                raise RuntimeError("Invalid id")

        self._add_argument(argument)

    def _add_argument(self, argument):
        if isinstance(argument, string_types):
            argument = StringHandler.prepare_incoming(argument)
        self._arguments.append(argument)
        self._sub_string = ""
        self._state = STATE_NEUTRAL



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/elements/lockable_button.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 2607 bytes
from ...base import EventObject, listenable_property, task
from .. import DOUBLE_CLICK_DELAY
from ..display import Renderable
from . import ButtonElement


class LockableButtonElementMixin(EventObject, Renderable):
    is_locked = listenable_property.managed(False)
    double_click_time = DOUBLE_CLICK_DELAY

    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self._locked_color = "DefaultButton.{}Locked".format(self.name.title().replace("_", ""))
        self._double_click_count = 0
        self._double_click_task = self._tasks.add(task.wait(self.double_click_time))
        self._double_click_task.kill()

    def reset(self):
        self._set_is_locked(False, do_notify=False)
        super().reset()

    def receive_value(self, value):
        if value:
            self._on_press()
        else:
            self._on_release()
        super().receive_value(value)

    def _on_press(self):
        self._set_is_locked(False)
        if not self._double_click_task.is_running:
            self._double_click_task.restart()
            self._double_click_count = 0

    def _on_release(self):
        if self._double_click_task.is_running:
            self._double_click_count += 1
            if self._double_click_count == 2:
                self._set_is_locked(True)
                self._double_click_task.kill()

    def _set_is_locked(self, is_locked, do_notify=True):
        if self.is_locked != is_locked:
            self.is_locked = is_locked
            if do_notify:
                self.notify(self.notifications.Element.button_lock, self.name, is_locked)

    def _set_skin_light(self, value):
        if self.is_locked:
            value = self._locked_color
        super()._set_skin_light(value)


class LockableButtonElement(LockableButtonElementMixin, ButtonElement):
    pass



from __future__ import annotations

from typing import Any, Callable


class FullVelocity:
    @property
    def enabled(self) -> Any:
        pass


class MidiRemoteScript:
    def handle(self) -> int:
        """
        handle( (MidiRemoteScript)arg1) -> int :

            C++ signature :
                int handle(TMidiRemoteScript*)
        """

    def instance_identifier(self) -> int:
        """
        instance_identifier( (MidiRemoteScript)arg1) -> int :

            C++ signature :
                int instance_identifier(TMidiRemoteScript*)
        """

    def log_message(self, arg2: object) -> None:
        """
        log_message( (MidiRemoteScript)arg1, (object)arg2) -> None :

            C++ signature :
                void log_message(TMidiRemoteScript*,TString)
        """

    def preferences(self, arg2: object) -> Preferences:
        """
        preferences( (MidiRemoteScript)arg1, (object)arg2) -> Preferences :

            C++ signature :
                TWeakPtr<APythonPreferenceEntry> preferences(TMidiRemoteScript*,TString)
        """

    def release_controlled_track(self) -> None:
        """
        release_controlled_track( (MidiRemoteScript)arg1) -> None :
            Disclaimer: The API functions/properties used for Push will be changed or removed and may not be available in the future.

            C++ signature :
                void release_controlled_track(TMidiRemoteScript*)
        """

    def request_rebuild_midi_map(self) -> None:
        """
        request_rebuild_midi_map( (MidiRemoteScript)arg1) -> None :

            C++ signature :
                void request_rebuild_midi_map(TMidiRemoteScript*)
        """

    def send_midi(self, arg2: object) -> None:
        """
        send_midi( (MidiRemoteScript)arg1, (tuple)arg2) -> None :

            C++ signature :
                void send_midi(TMidiRemoteScript*,boost::python::tuple)
        """

    def set_cc_translation(self, arg2: int, arg3: int, arg4: int, arg5: int) -> None:
        """
        set_cc_translation( (MidiRemoteScript)arg1, (int)arg2, (int)arg3, (int)arg4, (int)arg5) -> None :

            C++ signature :
                void set_cc_translation(TMidiRemoteScript*,int,int,int,int)
        """

    def set_controlled_track(self, arg2: object) -> None:
        """
        set_controlled_track( (MidiRemoteScript)arg1, (Track)arg2) -> None :
            Disclaimer: The API functions/properties used for Push will be changed or removed and may not be available in the future.

            C++ signature :
                void set_controlled_track(TMidiRemoteScript*,TTrackPyHandle)
        """

    def set_feedback_channels(self, arg2: object) -> None:
        """
        set_feedback_channels( (MidiRemoteScript)arg1, (list)arg2) -> None :
            Disclaimer: The API functions/properties used for Push will be changed or removed and may not be available in the future.

            C++ signature :
                void set_feedback_channels(TMidiRemoteScript*,boost::python::list)
        """

    def set_feedback_velocity(self, arg2: int) -> None:
        """
        set_feedback_velocity( (MidiRemoteScript)arg1, (int)arg2) -> None :
            Disclaimer: The API functions/properties used for Push will be changed or removed and may not be available in the future.

            C++ signature :
                void set_feedback_velocity(TMidiRemoteScript {lvalue},int)
        """

    def set_note_translation(self, arg2: int, arg3: int, arg4: int, arg5: int) -> None:
        """
        set_note_translation( (MidiRemoteScript)arg1, (int)arg2, (int)arg3, (int)arg4, (int)arg5) -> None :

            C++ signature :
                void set_note_translation(TMidiRemoteScript*,int,int,int,int)
        """

    def set_pad_translation(self, arg2: object) -> None:
        """
        set_pad_translation( (MidiRemoteScript)arg1, (tuple)arg2) -> None :

            C++ signature :
                void set_pad_translation(TMidiRemoteScript*,boost::python::tuple)
        """

    def set_session_highlight(
        self,
        track_offset: int,
        scene_offset: int,
        width: int,
        height: int,
        include_return_tracks: bool = False,
        include_rack_chains: bool = False,
    ) -> None:
        """
        set_session_highlight( (MidiRemoteScript)arg1, (int)track_offset, (int)scene_offset, (int)width, (int)height [, (bool)include_return_tracks=False [, (bool)include_rack_chains=False]]) -> None :

            C++ signature :
                void set_session_highlight(TMidiRemoteScript*,int,int,int,int [,bool=False [,bool=False]])
        """

    def show_message(self, arg2: object) -> None:
        """
        show_message( (MidiRemoteScript)arg1, (object)arg2) -> None :

            C++ signature :
                void show_message(TMidiRemoteScript*,TString)
        """

    def song(self) -> Song:
        """
        song( (MidiRemoteScript)arg1) -> Song :

            C++ signature :
                TWeakPtr<TPyHandle<ASong>> song(TMidiRemoteScript*)
        """

    def toggle_lock(self) -> None:
        """
        toggle_lock( (MidiRemoteScript)arg1) -> None :

            C++ signature :
                void toggle_lock(TMidiRemoteScript*)
        """

    def update_locks(self) -> None:
        """
        update_locks( (MidiRemoteScript)arg1) -> None :

            C++ signature :
                void update_locks(TMidiRemoteScript*)
        """

    @property
    def full_velocity(self) -> Any:
        """
        Disclaimer: The API functions/properties used for Push will be changed or removed and may not be available in the future.
        """

    @property
    def note_repeat(self) -> Any:
        """
        Disclaimer: The API functions/properties used for Push will be changed or removed and may not be available in the future.
        """

    @property
    def playhead(self) -> Any:
        """
        Disclaimer: The API functions/properties used for Push will be changed or removed and may not be available in the future.
        """

    @property
    def velocity_levels(self) -> Any:
        """
        Disclaimer: The API functions/properties used for Push will be changed or removed and may not be available in the future.
        """


class MutableVector:
    """
    A vector of things that is mutable and creatable from Python
    """

    def append(self, arg2: object) -> None:
        """
        append( (MutableVector)arg1, (object)arg2) -> None :

            C++ signature :
                void append(std::__1::vector<ableton::estd::variant<int, float, TPush2Visualisation::TVisualisationColor, std::__1::basic_string<wchar_t, std::__1::char_traits<wchar_t>, std::__1::allocator<wchar_t>>>, std::__1::allocator<ableton::estd::variant<int, float, TPush2Visualisation::TVisualisationColor, std::__1::basic_string<wchar_t, std::__1::char_traits<wchar_t>, std::__1::allocator<wchar_t>>>>> {lvalue},boost::python::api::object)
        """

    def extend(self, arg2: object) -> None:
        """
        extend( (MutableVector)arg1, (object)arg2) -> None :

            C++ signature :
                void extend(std::__1::vector<ableton::estd::variant<int, float, TPush2Visualisation::TVisualisationColor, std::__1::basic_string<wchar_t, std::__1::char_traits<wchar_t>, std::__1::allocator<wchar_t>>>, std::__1::allocator<ableton::estd::variant<int, float, TPush2Visualisation::TVisualisationColor, std::__1::basic_string<wchar_t, std::__1::char_traits<wchar_t>, std::__1::allocator<wchar_t>>>>> {lvalue},boost::python::api::object)
        """


class NoteRepeat:
    @property
    def enabled(self) -> Any:
        pass

    @property
    def repeat_rate(self) -> Any:
        pass


class Playhead:
    def set_feedback_channels(self, arg2: object) -> None:
        """
        set_feedback_channels( (Playhead)arg1, (list)arg2) -> None :
            Set the list of channels the notes representing playhead steps will be sent on.

            C++ signature :
                void set_feedback_channels(ATimeToMidiModule {lvalue},boost::python::list)
        """

    @property
    def clip(self) -> Clip:
        """
        The associated clip
        """

    @property
    def notes(self) -> Any:
        """
        Sequence of midi notes that represent the steps of the playhead
        """

    @property
    def start_time(self) -> Any:
        """
        The time at which the first midi note will be sent. It is relative to the clip.
        """

    @property
    def step_length(self) -> Any:
        """
        Time between each sent midi note.
        """

    @property
    def track(self) -> Track:
        """
        The track of the associated clip
        """

    @property
    def velocity(self) -> Any:
        """
        The velocity that is used for every sent note
        """

    @property
    def wrap_around(self) -> Any:
        """
        Enables/Disables wrapping of the playhead.
        """


class Preferences:
    def set_serializer(self, arg2: object) -> None:
        """
        set_serializer( (Preferences)arg1, (object)arg2) -> None :

            C++ signature :
                void set_serializer(APythonPreferenceEntry {lvalue},boost::python::api::object)
        """


class Push2MidiRemoteScript:
    def handle(self) -> int:
        """
        handle( (MidiRemoteScript)arg1) -> int :

            C++ signature :
                int handle(TMidiRemoteScript*)
        """

    def instance_identifier(self) -> int:
        """
        instance_identifier( (MidiRemoteScript)arg1) -> int :

            C++ signature :
                int instance_identifier(TMidiRemoteScript*)
        """

    def launch_external_process(self) -> None:
        """
        launch_external_process( (Push2MidiRemoteScript)arg1) -> None :
            Disclaimer: The API functions/properties used for Push and Push 2 will be changed or removed and may not be available in the future.

            C++ signature :
                void launch_external_process(TPush2MidiRemoteScript {lvalue})
        """

    def log_message(self, arg2: object) -> None:
        """
        log_message( (MidiRemoteScript)arg1, (object)arg2) -> None :

            C++ signature :
                void log_message(TMidiRemoteScript*,TString)
        """

    def preferences(self, arg2: object) -> Preferences:
        """
        preferences( (MidiRemoteScript)arg1, (object)arg2) -> Preferences :

            C++ signature :
                TWeakPtr<APythonPreferenceEntry> preferences(TMidiRemoteScript*,TString)
        """

    def process_connected(self) -> bool:
        """
        process_connected( (Push2MidiRemoteScript)arg1) -> bool :
            Disclaimer: The API functions/properties used for Push and Push 2 will be changed or removed and may not be available in the future.

            C++ signature :
                bool process_connected(TPush2MidiRemoteScript {lvalue})
        """

    def release_controlled_track(self) -> None:
        """
        release_controlled_track( (MidiRemoteScript)arg1) -> None :
            Disclaimer: The API functions/properties used for Push will be changed or removed and may not be available in the future.

            C++ signature :
                void release_controlled_track(TMidiRemoteScript*)
        """

    def request_rebuild_midi_map(self) -> None:
        """
        request_rebuild_midi_map( (MidiRemoteScript)arg1) -> None :

            C++ signature :
                void request_rebuild_midi_map(TMidiRemoteScript*)
        """

    def send_midi(self, arg2: object) -> None:
        """
        send_midi( (MidiRemoteScript)arg1, (tuple)arg2) -> None :

            C++ signature :
                void send_midi(TMidiRemoteScript*,boost::python::tuple)
        """

    def send_model_update(self, arg2: object) -> None:
        """
        send_model_update( (Push2MidiRemoteScript)arg1, (object)arg2) -> None :
            Disclaimer: The API functions/properties used for Push and Push 2 will be changed or removed and may not be available in the future.

            C++ signature :
                void send_model_update(TPush2MidiRemoteScript {lvalue},TArray<unsigned char>)
        """

    def set_cc_translation(self, arg2: int, arg3: int, arg4: int, arg5: int) -> None:
        """
        set_cc_translation( (MidiRemoteScript)arg1, (int)arg2, (int)arg3, (int)arg4, (int)arg5) -> None :

            C++ signature :
                void set_cc_translation(TMidiRemoteScript*,int,int,int,int)
        """

    def set_controlled_track(self, arg2: object) -> None:
        """
        set_controlled_track( (MidiRemoteScript)arg1, (Track)arg2) -> None :
            Disclaimer: The API functions/properties used for Push will be changed or removed and may not be available in the future.

            C++ signature :
                void set_controlled_track(TMidiRemoteScript*,TTrackPyHandle)
        """

    def set_feedback_channels(self, arg2: object) -> None:
        """
        set_feedback_channels( (MidiRemoteScript)arg1, (list)arg2) -> None :
            Disclaimer: The API functions/properties used for Push will be changed or removed and may not be available in the future.

            C++ signature :
                void set_feedback_channels(TMidiRemoteScript*,boost::python::list)
        """

    def set_feedback_velocity(self, arg2: int) -> None:
        """
        set_feedback_velocity( (MidiRemoteScript)arg1, (int)arg2) -> None :
            Disclaimer: The API functions/properties used for Push will be changed or removed and may not be available in the future.

            C++ signature :
                void set_feedback_velocity(TMidiRemoteScript {lvalue},int)
        """

    def set_firmware_version(self, arg2: float) -> None:
        """
        set_firmware_version( (PushMidiRemoteScript)arg1, (float)arg2) -> None :

            C++ signature :
                void set_firmware_version(TPushMidiRemoteScript {lvalue},float)
        """

    def set_note_translation(self, arg2: int, arg3: int, arg4: int, arg5: int) -> None:
        """
        set_note_translation( (MidiRemoteScript)arg1, (int)arg2, (int)arg3, (int)arg4, (int)arg5) -> None :

            C++ signature :
                void set_note_translation(TMidiRemoteScript*,int,int,int,int)
        """

    def set_pad_translation(self, arg2: object) -> None:
        """
        set_pad_translation( (MidiRemoteScript)arg1, (tuple)arg2) -> None :

            C++ signature :
                void set_pad_translation(TMidiRemoteScript*,boost::python::tuple)
        """

    def set_session_highlight(
        self,
        track_offset: int,
        scene_offset: int,
        width: int,
        height: int,
        include_return_tracks: bool = False,
        include_rack_chains: bool = False,
    ) -> None:
        """
        set_session_highlight( (MidiRemoteScript)arg1, (int)track_offset, (int)scene_offset, (int)width, (int)height [, (bool)include_return_tracks=False [, (bool)include_rack_chains=False]]) -> None :

            C++ signature :
                void set_session_highlight(TMidiRemoteScript*,int,int,int,int [,bool=False [,bool=False]])
        """

    def show_message(self, arg2: object) -> None:
        """
        show_message( (MidiRemoteScript)arg1, (object)arg2) -> None :

            C++ signature :
                void show_message(TMidiRemoteScript*,TString)
        """

    def song(self) -> Song:
        """
        song( (MidiRemoteScript)arg1) -> Song :

            C++ signature :
                TWeakPtr<TPyHandle<ASong>> song(TMidiRemoteScript*)
        """

    def toggle_lock(self) -> None:
        """
        toggle_lock( (MidiRemoteScript)arg1) -> None :

            C++ signature :
                void toggle_lock(TMidiRemoteScript*)
        """

    def update_locks(self) -> None:
        """
        update_locks( (MidiRemoteScript)arg1) -> None :

            C++ signature :
                void update_locks(TMidiRemoteScript*)
        """

    @property
    def full_velocity(self) -> Any:
        """
        Disclaimer: The API functions/properties used for Push will be changed or removed and may not be available in the future.
        """

    @property
    def note_repeat(self) -> Any:
        """
        Disclaimer: The API functions/properties used for Push will be changed or removed and may not be available in the future.
        """

    @property
    def playhead(self) -> Any:
        """
        Disclaimer: The API functions/properties used for Push will be changed or removed and may not be available in the future.
        """

    @property
    def real_time_mapper(self) -> Any:
        """
        Disclaimer: The API functions/properties used for Push and Push 2 will be changed or removed and may not be available in the future.
        """

    @property
    def velocity_levels(self) -> Any:
        """
        Disclaimer: The API functions/properties used for Push will be changed or removed and may not be available in the future.
        """


class Push2ProcessState:
    """
    Class that represent an enumeration of values for Push2ProcessState
    The state the external process is in.
    """

    initialized = 0
    launching = 1
    running = 2
    connected = 3
    died = 4
    terminated = 5
    killed = 6
    wait_for_termination = 7
    defunct_process_wait_for_termination = 8
    defunct_process_terminated = 9
    defunct_process_killed = 10


class PushMidiRemoteScript:
    def handle(self) -> int:
        """
        handle( (MidiRemoteScript)arg1) -> int :

            C++ signature :
                int handle(TMidiRemoteScript*)
        """

    def instance_identifier(self) -> int:
        """
        instance_identifier( (MidiRemoteScript)arg1) -> int :

            C++ signature :
                int instance_identifier(TMidiRemoteScript*)
        """

    def log_message(self, arg2: object) -> None:
        """
        log_message( (MidiRemoteScript)arg1, (object)arg2) -> None :

            C++ signature :
                void log_message(TMidiRemoteScript*,TString)
        """

    def preferences(self, arg2: object) -> Preferences:
        """
        preferences( (MidiRemoteScript)arg1, (object)arg2) -> Preferences :

            C++ signature :
                TWeakPtr<APythonPreferenceEntry> preferences(TMidiRemoteScript*,TString)
        """

    def release_controlled_track(self) -> None:
        """
        release_controlled_track( (MidiRemoteScript)arg1) -> None :
            Disclaimer: The API functions/properties used for Push will be changed or removed and may not be available in the future.

            C++ signature :
                void release_controlled_track(TMidiRemoteScript*)
        """

    def request_rebuild_midi_map(self) -> None:
        """
        request_rebuild_midi_map( (MidiRemoteScript)arg1) -> None :

            C++ signature :
                void request_rebuild_midi_map(TMidiRemoteScript*)
        """

    def send_midi(self, arg2: object) -> None:
        """
        send_midi( (MidiRemoteScript)arg1, (tuple)arg2) -> None :

            C++ signature :
                void send_midi(TMidiRemoteScript*,boost::python::tuple)
        """

    def set_cc_translation(self, arg2: int, arg3: int, arg4: int, arg5: int) -> None:
        """
        set_cc_translation( (MidiRemoteScript)arg1, (int)arg2, (int)arg3, (int)arg4, (int)arg5) -> None :

            C++ signature :
                void set_cc_translation(TMidiRemoteScript*,int,int,int,int)
        """

    def set_controlled_track(self, arg2: object) -> None:
        """
        set_controlled_track( (MidiRemoteScript)arg1, (Track)arg2) -> None :
            Disclaimer: The API functions/properties used for Push will be changed or removed and may not be available in the future.

            C++ signature :
                void set_controlled_track(TMidiRemoteScript*,TTrackPyHandle)
        """

    def set_feedback_channels(self, arg2: object) -> None:
        """
        set_feedback_channels( (MidiRemoteScript)arg1, (list)arg2) -> None :
            Disclaimer: The API functions/properties used for Push will be changed or removed and may not be available in the future.

            C++ signature :
                void set_feedback_channels(TMidiRemoteScript*,boost::python::list)
        """

    def set_feedback_velocity(self, arg2: int) -> None:
        """
        set_feedback_velocity( (MidiRemoteScript)arg1, (int)arg2) -> None :
            Disclaimer: The API functions/properties used for Push will be changed or removed and may not be available in the future.

            C++ signature :
                void set_feedback_velocity(TMidiRemoteScript {lvalue},int)
        """

    def set_firmware_version(self, arg2: float) -> None:
        """
        set_firmware_version( (PushMidiRemoteScript)arg1, (float)arg2) -> None :

            C++ signature :
                void set_firmware_version(TPushMidiRemoteScript {lvalue},float)
        """

    def set_note_translation(self, arg2: int, arg3: int, arg4: int, arg5: int) -> None:
        """
        set_note_translation( (MidiRemoteScript)arg1, (int)arg2, (int)arg3, (int)arg4, (int)arg5) -> None :

            C++ signature :
                void set_note_translation(TMidiRemoteScript*,int,int,int,int)
        """

    def set_pad_translation(self, arg2: object) -> None:
        """
        set_pad_translation( (MidiRemoteScript)arg1, (tuple)arg2) -> None :

            C++ signature :
                void set_pad_translation(TMidiRemoteScript*,boost::python::tuple)
        """

    def set_session_highlight(
        self,
        track_offset: int,
        scene_offset: int,
        width: int,
        height: int,
        include_return_tracks: bool = False,
        include_rack_chains: bool = False,
    ) -> None:
        """
        set_session_highlight( (MidiRemoteScript)arg1, (int)track_offset, (int)scene_offset, (int)width, (int)height [, (bool)include_return_tracks=False [, (bool)include_rack_chains=False]]) -> None :

            C++ signature :
                void set_session_highlight(TMidiRemoteScript*,int,int,int,int [,bool=False [,bool=False]])
        """

    def show_message(self, arg2: object) -> None:
        """
        show_message( (MidiRemoteScript)arg1, (object)arg2) -> None :

            C++ signature :
                void show_message(TMidiRemoteScript*,TString)
        """

    def song(self) -> Song:
        """
        song( (MidiRemoteScript)arg1) -> Song :

            C++ signature :
                TWeakPtr<TPyHandle<ASong>> song(TMidiRemoteScript*)
        """

    def toggle_lock(self) -> None:
        """
        toggle_lock( (MidiRemoteScript)arg1) -> None :

            C++ signature :
                void toggle_lock(TMidiRemoteScript*)
        """

    def update_locks(self) -> None:
        """
        update_locks( (MidiRemoteScript)arg1) -> None :

            C++ signature :
                void update_locks(TMidiRemoteScript*)
        """

    @property
    def full_velocity(self) -> Any:
        """
        Disclaimer: The API functions/properties used for Push will be changed or removed and may not be available in the future.
        """

    @property
    def note_repeat(self) -> Any:
        """
        Disclaimer: The API functions/properties used for Push will be changed or removed and may not be available in the future.
        """

    @property
    def playhead(self) -> Any:
        """
        Disclaimer: The API functions/properties used for Push will be changed or removed and may not be available in the future.
        """

    @property
    def velocity_levels(self) -> Any:
        """
        Disclaimer: The API functions/properties used for Push will be changed or removed and may not be available in the future.
        """


class RealTimeMapper:
    def attach_object(self, lom_object: object, channel_type: object) -> tuple:
        """
        attach_object( (RealTimeMapper)self, (object)lom_object, (object)channel_type) -> tuple :
            Attaches the provided LOM object to a suitable channel and returns that channel's id and LOM object id, each as a string. Returns an empty channel id string if no suitable channel is found.Calling this multiple times for the same object and channel type will attach anew channel each time! Each of these channels needs to be detached separately.

            C++ signature :
                boost::python::tuple attach_object(IPythonRealTimeMapper {lvalue},boost::python::api::object,TString)
        """

    def detach_channel(self, channel_id: object) -> None:
        """
        detach_channel( (RealTimeMapper)self, (object)channel_id) -> None :
            Detaches the channel with the provided id if it is attached to a lom_object.

            C++ signature :
                void detach_channel(IPythonRealTimeMapper {lvalue},TString)
        """

    METER_POOLSIZE = 9

    @property
    def client_id(self) -> Any:
        """
        The id used to identify the inter-process connection.
        """

    @property
    def device_visualisation(self) -> Any:
        """
        Visualisation object for device, or None if there isn't one
        """


class RgbaColor:
    @property
    def alpha(self) -> Any:
        pass

    @property
    def blue(self) -> Any:
        pass

    @property
    def green(self) -> Any:
        pass

    @property
    def red(self) -> Any:
        pass


class VelocityLevels:
    def add_last_played_level_listener(self, listener: Callable) -> None:
        """
        add_last_played_level_listener( (VelocityLevels)arg1, (object)arg2) -> None :

            C++ signature :
                void add_last_played_level_listener(AVelocityLevelsModule {lvalue},boost::python::api::object)
        """

    def last_played_level_has_listener(self, listener: Callable) -> bool:
        """
        last_played_level_has_listener( (VelocityLevels)arg1, (object)arg2) -> bool :

            C++ signature :
                bool last_played_level_has_listener(AVelocityLevelsModule {lvalue},boost::python::api::object)
        """

    def remove_last_played_level_listener(self, listener: Callable) -> None:
        """
        remove_last_played_level_listener( (VelocityLevels)arg1, (object)arg2) -> None :

            C++ signature :
                void remove_last_played_level_listener(AVelocityLevelsModule {lvalue},boost::python::api::object)
        """

    @property
    def enabled(self) -> Any:
        pass

    @property
    def last_played_level(self) -> Any:
        pass

    @property
    def levels(self) -> Any:
        pass

    @property
    def notes(self) -> Any:
        pass

    @property
    def source_channel(self) -> Any:
        pass

    @property
    def target_channel(self) -> Any:
        pass

    @property
    def target_note(self) -> Any:
        pass


class Visualisation:
    def get_view_data(self) -> VisualisationViewData:
        """
        get_view_data( (Visualisation)self) -> VisualisationViewData :
            Returns a copy of the view data of the visualisation

            C++ signature :
                std::__1::unordered_map<std::__1::basic_string<wchar_t, std::__1::char_traits<wchar_t>, std::__1::allocator<wchar_t>>, ableton::estd::variant<ableton::estd::variant<int, float, TPush2Visualisation::TVisualisationColor, std::__1::basic_string<wchar_t, std::__1::char_traits<wchar_t>, std::__1::allocator<wchar_t>>>, std::__1::vector<ableton::estd::variant<int, float, TPush2Visualisation::TVisualisationColor, std::__1::basic_string<wchar_t, std::__1::char_traits<wchar_t>, std::__1::allocator<wchar_t>>>, std::__1::allocator<ableton::estd::variant<int, float, TPush2Visualisation::TVisualisationColor, std::__1::basic_string<wchar_t, std::__1::char_traits<wchar_t>, std::__1::allocator<wchar_t>>>>>>, std::__1::hash<std::__1::basic_string<wchar_t, std::__1::char_traits<wchar_t>, std::__1::allocator<wchar_t>>>, std::__1::equal_to<std::__1::basic_string<wchar_t, std::__1::char_traits<wchar_t>, std::__1::allocator<wchar_t>>>, std::__1::allocator<std::__1::pair<std::__1::basic_string<wchar_t, std::__1::char_traits<wchar_t>, std::__1::allocator<wchar_t>> const, ableton::estd::variant<ableton::estd::variant<int, float, TPush2Visualisation::TVisualisationColor, std::__1::basic_string<wchar_t, std::__1::char_traits<wchar_t>, std::__1::allocator<wchar_t>>>, std::__1::vector<ableton::estd::variant<int, float, TPush2Visualisation::TVisualisationColor, std::__1::basic_string<wchar_t, std::__1::char_traits<wchar_t>, std::__1::allocator<wchar_t>>>, std::__1::allocator<ableton::estd::variant<int, float, TPush2Visualisation::TVisualisationColor, std::__1::basic_string<wchar_t, std::__1::char_traits<wchar_t>, std::__1::allocator<wchar_t>>>>>>>>> get_view_data(TPush2Visualisation {lvalue})
        """

    def set_view_data(self, view_data: object) -> None:
        """
        set_view_data( (Visualisation)self, (VisualisationViewData)view_data) -> None :
            Copies given view data to the visualisation

            C++ signature :
                void set_view_data(TPush2Visualisation {lvalue},std::__1::unordered_map<std::__1::basic_string<wchar_t, std::__1::char_traits<wchar_t>, std::__1::allocator<wchar_t>>, ableton::estd::variant<ableton::estd::variant<int, float, TPush2Visualisation::TVisualisationColor, std::__1::basic_string<wchar_t, std::__1::char_traits<wchar_t>, std::__1::allocator<wchar_t>>>, std::__1::vector<ableton::estd::variant<int, float, TPush2Visualisation::TVisualisationColor, std::__1::basic_string<wchar_t, std::__1::char_traits<wchar_t>, std::__1::allocator<wchar_t>>>, std::__1::allocator<ableton::estd::variant<int, float, TPush2Visualisation::TVisualisationColor, std::__1::basic_string<wchar_t, std::__1::char_traits<wchar_t>, std::__1::allocator<wchar_t>>>>>>, std::__1::hash<std::__1::basic_string<wchar_t, std::__1::char_traits<wchar_t>, std::__1::allocator<wchar_t>>>, std::__1::equal_to<std::__1::basic_string<wchar_t, std::__1::char_traits<wchar_t>, std::__1::allocator<wchar_t>>>, std::__1::allocator<std::__1::pair<std::__1::basic_string<wchar_t, std::__1::char_traits<wchar_t>, std::__1::allocator<wchar_t>> const, ableton::estd::variant<ableton::estd::variant<int, float, TPush2Visualisation::TVisualisationColor, std::__1::basic_string<wchar_t, std::__1::char_traits<wchar_t>, std::__1::allocator<wchar_t>>>, std::__1::vector<ableton::estd::variant<int, float, TPush2Visualisation::TVisualisationColor, std::__1::basic_string<wchar_t, std::__1::char_traits<wchar_t>, std::__1::allocator<wchar_t>>>, std::__1::allocator<ableton::estd::variant<int, float, TPush2Visualisation::TVisualisationColor, std::__1::basic_string<wchar_t, std::__1::char_traits<wchar_t>, std::__1::allocator<wchar_t>>>>>>>>>)
        """


class VisualisationViewData:
    def clear(self) -> None:
        """
        clear( (VisualisationViewData)arg1) -> None :

            C++ signature :
                void clear(std::__1::unordered_map<std::__1::basic_string<wchar_t, std::__1::char_traits<wchar_t>, std::__1::allocator<wchar_t>>, ableton::estd::variant<ableton::estd::variant<int, float, TPush2Visualisation::TVisualisationColor, std::__1::basic_string<wchar_t, std::__1::char_traits<wchar_t>, std::__1::allocator<wchar_t>>>, std::__1::vector<ableton::estd::variant<int, float, TPush2Visualisation::TVisualisationColor, std::__1::basic_string<wchar_t, std::__1::char_traits<wchar_t>, std::__1::allocator<wchar_t>>>, std::__1::allocator<ableton::estd::variant<int, float, TPush2Visualisation::TVisualisationColor, std::__1::basic_string<wchar_t, std::__1::char_traits<wchar_t>, std::__1::allocator<wchar_t>>>>>>, std::__1::hash<std::__1::basic_string<wchar_t, std::__1::char_traits<wchar_t>, std::__1::allocator<wchar_t>>>, std::__1::equal_to<std::__1::basic_string<wchar_t, std::__1::char_traits<wchar_t>, std::__1::allocator<wchar_t>>>, std::__1::allocator<std::__1::pair<std::__1::basic_string<wchar_t, std::__1::char_traits<wchar_t>, std::__1::allocator<wchar_t>> const, ableton::estd::variant<ableton::estd::variant<int, float, TPush2Visualisation::TVisualisationColor, std::__1::basic_string<wchar_t, std::__1::char_traits<wchar_t>, std::__1::allocator<wchar_t>>>, std::__1::vector<ableton::estd::variant<int, float, TPush2Visualisation::TVisualisationColor, std::__1::basic_string<wchar_t, std::__1::char_traits<wchar_t>, std::__1::allocator<wchar_t>>>, std::__1::allocator<ableton::estd::variant<int, float, TPush2Visualisation::TVisualisationColor, std::__1::basic_string<wchar_t, std::__1::char_traits<wchar_t>, std::__1::allocator<wchar_t>>>>>>>>> {lvalue})
        """



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/default_skin.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 9123 bytes
from .colors import BasicColors
from .skin import Skin, merge_skins


def create_skin(skin=None, colors=None):
    skins = [default_skin]
    if skin:
        skins.append(Skin(skin))
    if colors:
        skins.append(Skin(colors))
    return merge_skins(*skins)


class DefaultColors:
    class DefaultButton:
        On = BasicColors.ON
        Off = BasicColors.OFF
        Disabled = BasicColors.OFF

    class TargetTrack:
        LockOn = BasicColors.ON
        LockOff = BasicColors.OFF

    class Transport:
        PlayOn = BasicColors.ON
        PlayOff = BasicColors.OFF
        StopOn = BasicColors.ON
        StopOff = BasicColors.OFF
        AutomationArmOn = BasicColors.ON
        AutomationArmOff = BasicColors.OFF
        LoopOn = BasicColors.ON
        LoopOff = BasicColors.OFF
        MetronomeOn = BasicColors.ON
        MetronomeOff = BasicColors.OFF
        PunchOn = BasicColors.ON
        PunchOff = BasicColors.OFF
        TapTempoPressed = BasicColors.ON
        TapTempo = BasicColors.OFF
        NudgePressed = BasicColors.ON
        Nudge = BasicColors.OFF
        SeekPressed = BasicColors.ON
        Seek = BasicColors.OFF
        CanReEnableAutomation = BasicColors.ON
        CanCaptureMidi = BasicColors.ON
        CanJumpToCue = BasicColors.ON
        CannotJumpToCue = BasicColors.OFF
        SetCuePressed = BasicColors.ON
        SetCue = BasicColors.OFF
        RecordQuantizeOn = BasicColors.ON
        RecordQuantizeOff = BasicColors.OFF

    class Recording:
        ArrangementRecordOn = BasicColors.ON
        ArrangementRecordOff = BasicColors.OFF
        ArrangementOverdubOn = BasicColors.ON
        ArrangementOverdubOff = BasicColors.OFF
        SessionRecordOn = BasicColors.ON
        SessionRecordTransition = BasicColors.ON
        SessionRecordOff = BasicColors.OFF
        SessionOverdubOn = BasicColors.ON
        SessionOverdubOff = BasicColors.OFF
        NewPressed = BasicColors.ON
        New = BasicColors.OFF

    class UndoRedo:
        UndoPressed = BasicColors.ON
        Undo = BasicColors.OFF
        RedoPressed = BasicColors.ON
        Redo = BasicColors.OFF

    class ViewControl:
        TrackPressed = BasicColors.ON
        Track = BasicColors.ON
        ScenePressed = BasicColors.ON
        Scene = BasicColors.ON

    class ViewToggle:
        SessionOn = BasicColors.ON
        SessionOff = BasicColors.OFF
        DetailOn = BasicColors.ON
        DetailOff = BasicColors.OFF
        ClipOn = BasicColors.ON
        ClipOff = BasicColors.OFF
        BrowserOn = BasicColors.ON
        BrowserOff = BasicColors.OFF

    class Mixer:
        ArmOn = BasicColors.ON
        ArmOff = BasicColors.OFF
        ImplicitArmOn = BasicColors.ON
        MuteOn = BasicColors.ON
        MuteOff = BasicColors.OFF
        SoloOn = BasicColors.ON
        SoloOff = BasicColors.OFF
        Selected = BasicColors.ON
        NotSelected = BasicColors.OFF
        CrossfadeA = BasicColors.ON
        CrossfadeB = BasicColors.ON
        CrossfadeOff = BasicColors.OFF
        CycleSendIndexPressed = BasicColors.OFF
        CycleSendIndex = BasicColors.ON
        CycleSendIndexDisabled = BasicColors.OFF
        NoTrack = BasicColors.OFF

    class Session:
        Slot = BasicColors.OFF
        SlotRecordButton = BasicColors.OFF
        NoSlot = BasicColors.OFF
        ClipStopped = BasicColors.OFF
        ClipTriggeredPlay = BasicColors.ON
        ClipTriggeredRecord = BasicColors.ON
        ClipPlaying = BasicColors.ON
        ClipRecording = BasicColors.ON
        Scene = BasicColors.OFF
        SceneTriggered = BasicColors.ON
        NoScene = BasicColors.OFF
        StopClipTriggered = BasicColors.ON
        StopClip = BasicColors.OFF
        StopClipDisabled = BasicColors.OFF
        StopAllClipsPressed = BasicColors.ON
        StopAllClips = BasicColors.OFF
        NavigationPressed = BasicColors.ON
        Navigation = BasicColors.ON

    class Zooming:
        Selected = BasicColors.OFF
        Stopped = BasicColors.ON
        Playing = BasicColors.ON
        Empty = BasicColors.OFF

    class ClipActions:
        Delete = BasicColors.OFF
        DeletePressed = BasicColors.ON
        Double = BasicColors.OFF
        DoublePressed = BasicColors.ON
        Duplicate = BasicColors.OFF
        DuplicatePressed = BasicColors.ON
        Quantize = BasicColors.OFF
        QuantizedPressed = BasicColors.ON

    class Device:
        On = BasicColors.ON
        Off = BasicColors.OFF
        LockOn = BasicColors.ON
        LockOff = BasicColors.OFF
        NavigationPressed = BasicColors.ON
        Navigation = BasicColors.OFF

        class Bank:
            Selected = BasicColors.ON
            NotSelected = BasicColors.OFF
            NavigationPressed = BasicColors.ON
            Navigation = BasicColors.ON

    class Accent:
        On = BasicColors.ON
        Off = BasicColors.OFF

    class DrumGroup:
        PadEmpty = BasicColors.OFF
        PadFilled = BasicColors.OFF
        PadSelected = BasicColors.ON
        PadMuted = BasicColors.ON
        PadMutedSelected = BasicColors.ON
        PadSoloed = BasicColors.ON
        PadSoloedSelected = BasicColors.ON
        PadAction = BasicColors.ON
        ScrollPressed = BasicColors.ON
        Scroll = BasicColors.ON

    class SlicedSimpler:
        NoSlice = BasicColors.OFF
        SliceNotSelected = BasicColors.OFF
        SliceSelected = BasicColors.ON
        NextSlice = BasicColors.ON
        PadAction = BasicColors.ON
        ScrollPressed = BasicColors.ON
        Scroll = BasicColors.ON

    class NoteEditor:
        NoClip = BasicColors.OFF
        StepDisabled = BasicColors.OFF
        StepEmpty = BasicColors.OFF
        StepFilled = BasicColors.ON
        StepMuted = BasicColors.OFF

        class Resolution:
            Selected = BasicColors.ON
            NotSelected = BasicColors.OFF

    class LoopSelector:
        InsideLoopSelected = BasicColors.ON
        InsideLoop = BasicColors.OFF
        OutsideLoopSelected = BasicColors.ON
        OutsideLoop = BasicColors.OFF
        Playhead = BasicColors.OFF
        PlayheadRecord = BasicColors.OFF
        NavigationPressed = BasicColors.ON
        Navigation = BasicColors.OFF

    class Clipboard:
        Empty = BasicColors.OFF
        Filled = BasicColors.ON

    class Translation:
        class Channel:
            Selected = BasicColors.ON
            NotSelected = BasicColors.OFF


default_skin = Skin(DefaultColors)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Framework/MixerComponent.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 13747 bytes
from future.moves.itertools import zip_longest

from .ChannelStripComponent import ChannelStripComponent, release_control
from .CompoundComponent import CompoundComponent
from .SubjectSlot import subject_slot
from .Util import clamp


def turn_button_on_off(button, on=True):
    if button is not None:
        if on:
            button.turn_on()
        else:
            button.turn_off()


class MixerComponent(CompoundComponent):
    def __init__(self, num_tracks=0, num_returns=0, auto_name=False, invert_mute_feedback=False, *a, **k):
        (super().__init__)(*a, **k)
        self._track_offset = -1
        self._send_index = 0
        self._bank_up_button = None
        self._bank_down_button = None
        self._next_track_button = None
        self._prev_track_button = None
        self._prehear_volume_control = None
        self._crossfader_control = None
        self._send_controls = None
        self._channel_strips = []
        self._return_strips = []
        self._offset_can_start_after_tracks = False
        for index in range(num_tracks):
            strip = self._create_strip()
            self._channel_strips.append(strip)
            self.register_components(self._channel_strips[index])
            if invert_mute_feedback:
                strip.set_invert_mute_feedback(True)

        for index in range(num_returns):
            self._return_strips.append(self._create_strip())
            self.register_components(self._return_strips[index])

        self._master_strip = self._create_strip()
        self.register_components(self._master_strip)
        self._master_strip.set_track(self.song().master_track)
        self._selected_strip = self._create_strip()
        self.register_components(self._selected_strip)
        self.on_selected_track_changed()
        self.set_track_offset(0)
        if auto_name:
            self._auto_name()
        self._on_return_tracks_changed.subject = self.song()
        self._on_return_tracks_changed()

        def make_button_slot(name):
            return self.register_slot(None, getattr(self, f"_{name}_value"), "value")

        self._bank_up_button_slot = make_button_slot("bank_up")
        self._bank_down_button_slot = make_button_slot("bank_down")
        self._next_track_button_slot = make_button_slot("next_track")
        self._prev_track_button_slot = make_button_slot("prev_track")

    def disconnect(self):
        super().disconnect()
        release_control(self._prehear_volume_control)
        release_control(self._crossfader_control)
        self._bank_up_button = None
        self._bank_down_button = None
        self._next_track_button = None
        self._prev_track_button = None
        self._prehear_volume_control = None
        self._crossfader_control = None

    def _get_send_index(self):
        return self._send_index

    def _set_send_index(self, index):
        if index is None or 0 <= index < self.num_sends:
            if self._send_index != index:
                self._send_index = index
                self.set_send_controls(self._send_controls)
                self.on_send_index_changed()
        else:
            raise IndexError

    send_index = property(_get_send_index, _set_send_index)

    def on_send_index_changed(self):
        pass

    @property
    def num_sends(self):
        return len(self.song().return_tracks)

    def channel_strip(self, index):
        return self._channel_strips[index]

    def return_strip(self, index):
        return self._return_strips[index]

    def master_strip(self):
        return self._master_strip

    def selected_strip(self):
        return self._selected_strip

    def set_prehear_volume_control(self, control):
        release_control(self._prehear_volume_control)
        self._prehear_volume_control = control
        self.update()

    def set_crossfader_control(self, control):
        release_control(self._crossfader_control)
        self._crossfader_control = control
        self.update()

    def set_volume_controls(self, controls):
        for strip, control in zip_longest(self._channel_strips, controls or []):
            strip.set_volume_control(control)

    def set_pan_controls(self, controls):
        for strip, control in zip_longest(self._channel_strips, controls or []):
            strip.set_pan_control(control)

    def set_send_controls(self, controls):
        self._send_controls = controls
        for strip, control in zip_longest(self._channel_strips, controls or []):
            if self._send_index is None:
                strip.set_send_controls(None)
            else:
                strip.set_send_controls((None,) * self._send_index + (control,))

    def set_arm_buttons(self, buttons):
        for strip, button in zip_longest(self._channel_strips, buttons or []):
            strip.set_arm_button(button)

    def set_solo_buttons(self, buttons):
        for strip, button in zip_longest(self._channel_strips, buttons or []):
            strip.set_solo_button(button)

    def set_mute_buttons(self, buttons):
        for strip, button in zip_longest(self._channel_strips, buttons or []):
            strip.set_mute_button(button)

    def set_track_select_buttons(self, buttons):
        for strip, button in zip_longest(self._channel_strips, buttons or []):
            strip.set_select_button(button)

    def set_shift_button(self, button):
        for strip in self._channel_strips or []:
            strip.set_shift_button(button)

    def set_bank_buttons(self, up_button, down_button):
        do_update = False
        if up_button is not self._bank_up_button:
            do_update = True
            self._bank_up_button = up_button
            self._bank_up_button_slot.subject = up_button
        if down_button is not self._bank_down_button:
            do_update = True
            self._bank_down_button = down_button
            self._bank_down_button_slot.subject = down_button
        if do_update:
            self.on_track_list_changed()

    def set_select_buttons(self, next_button, prev_button):
        do_update = False
        if next_button is not self._next_track_button:
            do_update = True
            self._next_track_button = next_button
            self._next_track_button_slot.subject = next_button
        if prev_button is not self._prev_track_button:
            do_update = True
            self._prev_track_button = prev_button
            self._prev_track_button_slot.subject = prev_button
        if do_update:
            self.on_selected_track_changed()

    def set_track_offset(self, new_offset):
        if new_offset != self._track_offset:
            self._offset_can_start_after_tracks |= new_offset > len(self.tracks_to_use()) - 1
            self._track_offset = new_offset
            self._reassign_tracks()

    def on_enabled_changed(self):
        self.update()

    def on_track_list_changed(self):
        if not self._offset_can_start_after_tracks:
            self._track_offset = min(self._track_offset, len(self.tracks_to_use()) - 1)
        self._reassign_tracks()

    def on_selected_track_changed(self):
        selected_track = self.song().view.selected_track
        if self._selected_strip is not None:
            self._selected_strip.set_track(selected_track)
        if self.is_enabled():
            turn_button_on_off((self._next_track_button), on=(selected_track != self.song().master_track))
            turn_button_on_off((self._prev_track_button), on=(selected_track != self.song().visible_tracks[0]))

    @subject_slot("return_tracks")
    def _on_return_tracks_changed(self):
        num_sends = self.num_sends
        if self._send_index is not None:
            self.send_index = clamp(self._send_index, 0, num_sends - 1) if num_sends > 0 else None
        else:
            self.send_index = 0 if num_sends > 0 else None
        self.on_num_sends_changed()

    def on_num_sends_changed(self):
        pass

    def tracks_to_use(self):
        return self.song().visible_tracks

    def update(self):
        super().update()
        if self._allow_updates:
            master_track = self.song().master_track
            if self.is_enabled():
                if self._prehear_volume_control is not None:
                    self._prehear_volume_control.connect_to(master_track.mixer_device.cue_volume)
                if self._crossfader_control is not None:
                    self._crossfader_control.connect_to(master_track.mixer_device.crossfader)
            else:
                release_control(self._prehear_volume_control)
                release_control(self._crossfader_control)
                [
                    turn_button_on_off(x, on=False)
                    for x in [
                        self._bank_up_button,
                        self._bank_down_button,
                        self._next_track_button,
                        self._prev_track_button,
                    ]
                ]
        else:
            self._update_requests += 1

    def _reassign_tracks(self):
        tracks = self.tracks_to_use()
        returns = self.song().return_tracks
        num_strips = len(self._channel_strips)
        for index in range(num_strips):
            track_index = self._track_offset + index
            track = tracks[track_index] if len(tracks) > track_index else None
            self._channel_strips[index].set_track(track)

        for index in range(len(self._return_strips)):
            if len(returns) > index:
                self._return_strips[index].set_track(returns[index])
            else:
                self._return_strips[index].set_track(None)

        turn_button_on_off((self._bank_down_button), on=(self._track_offset > 0))
        turn_button_on_off((self._bank_up_button), on=(len(tracks) > self._track_offset + num_strips))

    def _create_strip(self):
        return ChannelStripComponent()

    def _bank_up_value(self, value):
        if self.is_enabled():
            if not (value != 0 or self._bank_up_button.is_momentary()):
                new_offset = self._track_offset + len(self._channel_strips)
                if len(self.tracks_to_use()) > new_offset:
                    self.set_track_offset(new_offset)

    def _bank_down_value(self, value):
        if self.is_enabled():
            if not (value != 0 or self._bank_down_button.is_momentary()):
                self.set_track_offset(max(0, self._track_offset - len(self._channel_strips)))

    def _next_track_value(self, value):
        if self.is_enabled():
            if not (value != 0 or self._next_track_button.is_momentary()):
                selected_track = self.song().view.selected_track
                all_tracks = (
                    tuple(self.song().visible_tracks) + tuple(self.song().return_tracks) + (self.song().master_track,)
                )
                if selected_track != all_tracks[-1]:
                    index = list(all_tracks).index(selected_track)
                    self.song().view.selected_track = all_tracks[index + 1]

    def _prev_track_value(self, value):
        if self.is_enabled():
            if not (value != 0 or self._prev_track_button.is_momentary()):
                selected_track = self.song().view.selected_track
                all_tracks = (
                    tuple(self.song().visible_tracks) + tuple(self.song().return_tracks) + (self.song().master_track,)
                )
                if selected_track != all_tracks[0]:
                    index = list(all_tracks).index(selected_track)
                    self.song().view.selected_track = all_tracks[index - 1]

    def _auto_name(self):
        self.name = "Mixer"
        self.master_strip().name = "Master_Channel_Strip"
        self.selected_strip().name = "Selected_Channel_Strip"
        for index, strip in enumerate(self._channel_strips):
            strip.name = "Channel_Strip_%d" % index



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/components/paginator.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 2492 bytes
from ...base import EventObject, depends, forward_property, listens
from .. import Component


class Paginator(EventObject):
    __events__ = ("page", "page_time", "page_length")
    can_change_page = NotImplemented
    page_length = NotImplemented
    page_time = NotImplemented


class NoteEditorPaginator(Component, Paginator):
    can_change_page = forward_property("_note_editor")("can_change_page")
    page_length = forward_property("_note_editor")("page_length")

    @depends(note_editor=None)
    def __init__(self, note_editor=None, *a, **k):
        (super().__init__)(*a, **k)
        self._note_editor = note_editor
        self._last_page_time = 0.0
        self._NoteEditorPaginator__on_page_length_changed.subject = note_editor
        self._NoteEditorPaginator__on_active_steps_changed.subject = note_editor

    @property
    def page_time(self):
        return self._note_editor.page_time

    @page_time.setter
    def page_time(self, time):
        can_change_page = self.can_change_page
        if can_change_page:
            if time != self._last_page_time:
                self._note_editor.page_time = time
                self._last_page_time = time
                self.notify_page()
                self.notify_page_time()

    def update(self):
        super().update()
        if self.is_enabled():
            self.notify_page_time()
            self.notify_page()
            self.notify_page_length()

    @listens("active_steps")
    def __on_active_steps_changed(self):
        if self.is_enabled():
            self.notify_page()

    @listens("page_length")
    def __on_page_length_changed(self):
        if self.is_enabled():
            self.notify_page()
            self.notify_page_length()



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/component_map.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 3731 bytes
from functools import partial

from . import Component
from .components import (
    AccentComponent,
    ActiveParameterComponent,
    ClipActionsComponent,
    DeviceComponent,
    DrumGroupComponent,
    MixerComponent,
    ModifierBackgroundComponent,
    RecordingComponent,
    SessionComponent,
    SessionNavigationComponent,
    SessionOverviewComponent,
    SimpleDeviceNavigationComponent,
    SlicedSimplerComponent,
    TranslatingBackgroundComponent,
    TransportComponent,
    UndoRedoComponent,
    ViewBasedRecordingComponent,
    ViewControlComponent,
    ViewToggleComponent,
)


class ComponentMap(dict):
    def __init__(self, specification, *a, **k):
        (super().__init__)(*a, **k)
        self._create_component_map(specification)

    def get(self, key, **_):
        return self.__getitem__(key)

    def __getitem__(self, key):
        component_or_factory = super().__getitem__(key)
        if isinstance(component_or_factory, Component):
            return component_or_factory
        component = component_or_factory(is_enabled=False)
        self[key] = component
        return component

    def __contains__(self, key):
        return isinstance(super().get(key), Component)

    def _create_component_map(self, specification):
        self["Accent"] = AccentComponent
        self["Active_Parameter"] = ActiveParameterComponent
        self["Clip_Actions"] = ClipActionsComponent
        self["Device"] = partial(
            DeviceComponent,
            bank_definitions=(specification.parameter_bank_definitions),
            bank_size=(specification.parameter_bank_size),
            continuous_parameter_sensitivity=(specification.continuous_parameter_sensitivity),
            quantized_parameter_sensitivity=(specification.quantized_parameter_sensitivity),
        )
        self["Device_Navigation"] = SimpleDeviceNavigationComponent
        self["Drum_Group"] = DrumGroupComponent
        self["Mixer"] = MixerComponent
        self["Modifier_Background"] = ModifierBackgroundComponent
        self["Recording"] = partial(RecordingComponent, recording_method_type=(specification.recording_method_type))
        self["Session"] = SessionComponent
        self["Session_Navigation"] = partial(
            SessionNavigationComponent,
            snap_track_offset=(specification.snap_track_offset),
        )
        self["Session_Overview"] = SessionOverviewComponent
        self["Sliced_Simpler"] = SlicedSimplerComponent
        self["Translating_Background"] = TranslatingBackgroundComponent
        self["Transport"] = TransportComponent
        self["Undo_Redo"] = UndoRedoComponent
        self["View_Based_Recording"] = partial(
            ViewBasedRecordingComponent,
            recording_method_type=(specification.recording_method_type),
        )
        self["View_Control"] = ViewControlComponent
        self["View_Toggle"] = ViewToggleComponent
        self.update(specification.component_map)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/elements/sysex_element.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 3025 bytes
from base.dependency import depends

from ableton.v2.base import old_hasattr

from ..input_control_element import MIDI_SYSEX_TYPE, InputControlElement


class SysexElement(InputControlElement):
    def __init__(self, send_message_generator=None, enquire_message=None, default_value=None, optimized=False, *a, **k):
        (super().__init__)(a, msg_type=MIDI_SYSEX_TYPE, **k)
        self._send_message_generator = send_message_generator
        self._enquire_message = enquire_message
        self._default_value = default_value
        self._optimized = optimized

    def message_map_mode(self):
        raise AssertionError("SysexElement doesn't support mapping.")

    def send_value(self, *arguments):
        message = (self._send_message_generator)(*arguments)
        self._do_send_value(message)

    def enquire_value(self):
        self.send_midi(self._enquire_message)

    def reset(self):
        if self._default_value is not None:
            self.send_value(self._default_value)

    def _do_send_value(self, message):
        if self._optimized:
            if message == self._last_sent_message or self.send_midi(message):
                self._last_sent_message = message
        else:
            self.send_midi(message)

    @property
    def _last_sent_value(self):
        if self._last_sent_message:
            return self._last_sent_message
        return -1


class ColorSysexElement(SysexElement):
    @depends(skin=None)
    def __init__(self, skin=None, *a, **k):
        (super().__init__)(*a, **k)
        self._skin = skin

    def set_light(self, value):
        color = None
        color = value if old_hasattr(value, "draw") else self._skin[value]
        color.draw(self)
